/*---------------------------------------------------------------------
Copyright Frédéric Legrand (2014)
  
This software is a computer program whose purpose is to solve Poisson
 equation with a finite difference method. 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,  
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".  

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.  

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 
 
The fact that you are presently reading this means that you have had 
knowledge of the CeCILL license and that you accept its terms.
 
-------------------------------------------------------------------------*/
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "poissonMG.h"
#ifdef OPENCL
#include <CL/cl.h>
#endif
#include <pthread.h>

/* conventions de notation : 
    i,j : indices d'un point sur la grille 0
    ig, jg : indice d'un point sur la grille en cours de traitement (ig = i*r)

*/

typedef struct {
    int index;
    int pw,ph; // puissances de 2 de la grille
    int r; // rapport à la première grille
    int g; // niveau de la grille 
    short width,height; // taille de la grille (nombre de points)
    short full_width, full_height; // taille de la grille complète (avec les rangées fictives)
    int size, full_size; // taille total des tableaux
    float dx,dy; // pas de la grille
    int jg_min; // début du domaine d'intégration pour la grille polaire
    float *G,*D,*H,*B,*C,*F,*U,*R;
    float *U1;
    float *FR; // facteur à appliquer au résidu (pour la réduction inter-grille)
    char *GT; // transfert intergrille à faire
    char *P; // tracé et coloriage des polygones
    unsigned char *M; // masque pour la représentation des objets
#ifdef OPENCL
    cl_mem gpu_G,gpu_D,gpu_H,gpu_B,gpu_C,gpu_F,gpu_U,gpu_V;
#endif
} grid_def;

typedef struct {
    char type;
    int pw_min, ph_min;
    int grid_levels;
    grid_def* fine_grid;
    float w,h;
    int (*neumann_function)(grid_def* grid, int igp, int jgp, int ig, int jg, int ign, int jgn, float source, float derivX, float derivY);
    int (*interface_function)(grid_def* grid, int igp, int jgp, int ig, int jg, int ign, int jgn, float source,float a1, float a2);
    int (*laplacien_function)(grid_def* grid);
    int (*source_function)(grid_def* grid, int ig, int jg, float s);
    void (*prolongation_function)(grid_def *grid_2h, grid_def *grid_h);
    void (*restriction_function)(grid_def *grid_2h, grid_def *grid_h);
    void (*iterations_function)(grid_def* grid, int niter, float w);
    void (*residu_function)(grid_def* grid);
} solver_def;

pthread_barrier_t threadBarrier;
pthread_cond_t condition_fin_iteration = PTHREAD_COND_INITIALIZER;
pthread_mutex_t mutex_fin_iteration = PTHREAD_MUTEX_INITIALIZER;
typedef struct {
    int niter;
    int jStart;
    int jEnd;
    grid_def* grid;
    float w1;
    float w;
} threadArg;



grid_def** grid_list;
solver_def solver;



int _laplacien(grid_def* grid) {
    int k;
    float dx2,dy2;
    float r;
    dx2=(grid->dx)*(grid->dx);
    dy2=(grid->dy)*(grid->dy);
    r = dy2/dx2;
    for (k=0; k<grid->full_size; k++) {
        grid->D[k] = grid->G[k] = r;
        grid->H[k] = grid->B[k] = 1.0;
        grid->C[k] = -2*(1+r);
        grid->F[k] = 0.0;
        grid->U[k] = 0.0;
        grid->FR[k] = 4.0;
    }
    return 0;
}




int _laplacien_axial(grid_def* grid) {
    int ig,jg,k;
    float dx2,dy2;
    float r;
    float h;
    h = grid->height;
    dx2=(grid->dx)*(grid->dx);
    dy2=(grid->dy)*(grid->dy);
    r = dy2/dx2;
    for (ig=0; ig<grid->width; ig++) {
        jg=0;
        k = (ig+1)+grid->full_width*(jg+1);
        grid->D[k] = grid->G[k] = r/8.0;
        grid->H[k] = 0.5;
        grid->B[k] = 0.0;
        grid->C[k] = -r/4.0-0.5;
        grid->F[k] = 0.0;
        grid->U[k] = 0.0;
        grid->FR[k] = 4.0;
        for (jg=1; jg<grid->height; jg++) {
            k = (ig+1)+grid->full_width*(jg+1);
            grid->D[k] = grid->G[k] = ((float)jg)/h*r;
            grid->H[k] = (jg+0.5)/h;
            grid->B[k] = (jg-0.5)/h;
            grid->C[k] = -2*((float)jg)/h*(1.0+r);
            grid->F[k] = 0.0;
            grid->U[k] = 0.0;
            grid->FR[k] = 4.0;
        }
        
    }
    return 0;
}

int _laplacien_axial_vect(grid_def* grid) {
    int ig,jg,k;
    float dx2,dy2;
    float r;
    dx2=(grid->dx)*(grid->dx);
    dy2=(grid->dy)*(grid->dy);
    r = dy2/dx2;
    for (ig=0; ig<grid->width; ig++) {
        jg=0;
        k = (ig+1)+grid->full_width*(jg+1);
        grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
        grid->C[k] = 1.0;
        grid->U[k] = grid->F[k] = 0;
        grid->FR[k] = 4.0;
        for (jg=1; jg<grid->height; jg++) {
            k = (ig+1)+grid->full_width*(jg+1);
            grid->D[k] = grid->G[k] = r;
            grid->H[k] = (jg+1.0)/(jg+0.5);
            grid->B[k] = (jg-1.0)/(jg-0.5);
            grid->C[k] = -(jg/(jg+0.5)+jg/(jg-0.5)+2*r);
            grid->F[k] = 0.0;
            grid->U[k] = 0.0;
            grid->FR[k] = 4.0;
        }
    }
    return 0;
}

int _laplacien_polar(grid_def* grid) {
    int ig,jg,k;
    float dtheta2;
    dtheta2 = grid->dx*grid->dx;
    for (jg=1; jg<grid->height; jg++) {
        for (ig=0; ig<grid->width; ig++) {
             k = (ig+1)+grid->full_width*(jg+1);
             grid->D[k] = grid->G[k] = 1.0;
             grid->H[k] = jg*(jg+0.5)*dtheta2;
             grid->B[k] = jg*(jg-0.5)*dtheta2;
             grid->C[k] = -(2*jg*jg*dtheta2+2);
             grid->F[k] = 0.0;
             grid->U[k] = 0.0;
             grid->FR[k] = 4.0;
        }
    }
    return 0;
}

int poissonMG_laplacien() {
    int g;
    for (g=0; g<solver.grid_levels; g++) {
        solver.laplacien_function(grid_list[g]);   
    }
    return 0;
}

int _source(grid_def* grid, int ig, int jg, float s) {
    int k;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    grid->F[k] = (grid->dy)*(grid->dy)*s;
    return 0;
}

int _source_axial(grid_def* grid, int ig, int jg, float s) {
    int k;
    float h;
    h = grid->height;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    if (jg==0) {
        grid->F[k] = grid->dy*grid->dy/8.0;   
    }
    else {
        grid->F[k] = ((float)jg)/h*grid->dy*grid->dy*s;
    }
    return 0;    
}

int _source_polar(grid_def* grid, int ig, int jg, float s) {
    int k;
    float dtheta2,dr2;
    dtheta2 = grid->dx*grid->dx;
    dr2 = grid->dy*grid->dy;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    if (jg==0) { // terme de source pour le centre
        grid->F[k] = 3.1415927*dr2/4*s;   
    }
    else {
        grid->F[k] = jg*jg*dr2*dtheta2*s;
    }
    return 0;    
}


int poissonMG_source(int i, int j, float s) {
   int g,err;
   for (g=0; g<solver.grid_levels; g++) {
        err = _source(grid_list[g],i*grid_list[g]->r,j*grid_list[g]->r,s);
        if (err!=0) return err;
    }
    return 0;    
}

int _source_polygon(grid_def* grid, int i_start, int j_start, intArray li_list, intArray lj_list, float value) {
    int k,nig,njg,ig,jg,ig_start,jg_start,id,jd;
    if ((lj_list.ndims!=1)||(lj_list.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (li_list.dims[0]!=lj_list.dims[0]) return POISSON_BAD_ARRAY_DIM;
    ig_start = i_start*grid->r;
    jg_start = j_start*grid->r;
    for (k=0; k<li_list.dims[0]; k++) {
        nig = li_list.data[k]*grid->r;
        njg = lj_list.data[k]*grid->r;
        if (nig!=0) {
               jd = 0;
               id = 1;
               if (nig<0) id = -1;
               if (njg!=0) return POISSON_BAD_POLYGON_DEFINITION;
               jg = jg_start;
               for (ig=ig_start; ig!=ig_start+nig+id; ig+=id) {
                   grid->P[ig+grid->width*jg] = MASK_SOURCE;
                   solver.source_function(grid,ig,jg,value);
               }
        }
        if (njg!=0) {
               id = 0;
               jd = 1;
               if (njg<0) jd = -1;
               if (nig!=0) return POISSON_BAD_POLYGON_DEFINITION;
               ig = ig_start;
               for (jg=jg_start; jg!=jg_start+njg+jd; jg+=jd) {
                   grid->P[ig+grid->width*jg] = MASK_SOURCE;
                   solver.source_function(grid,ig,jg,value);
               }
        }
        ig_start = ig;
        jg_start = jg;
    }
    return 0;  
}




int poissonMG_source_polygon(int i_start, int j_start, intArray li_list, intArray lj_list,float value) {
    int g,err;
    for (g=0; g<solver.grid_levels; g++) {
        err = _source_polygon(grid_list[g],i_start,j_start,li_list,lj_list,value);  
        if (err<0) return err;
    }
    return 0;  
}

int _source_rect(grid_def* grid, int cx, int cy, int width, int height, float s) {
    int ig,jg;
    for (ig=cx-width; ig<=cx+width; ig++) {
        for (jg=cy-height; jg<=cy+height; jg++) {
            grid->P[ig+grid->width*jg] = MASK_SOURCE;
            solver.source_function(grid,ig,jg,s);   
        }
    }
    return 0;
}




int poissonMG_source_rect(int cx, int cy, int width, int height, float s) {
    int g;
    for (g=0; g<solver.grid_levels; g++) {
        _source_rect(grid_list[g],cx*grid_list[g]->r,cy*grid_list[g]->r,width*grid_list[g]->r,height*grid_list[g]->r,s);   
    }
    return 0;   
}

int _dirichlet(grid_def* grid , int ig, int jg, float value) {
    int k;
    if ((ig<0)||(ig>grid->width)||(jg<0)||(jg>grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
    grid->C[k] = 1;
    grid->U[k] = grid->F[k] = value;
    grid->GT[k] = 0.0;
    return 0;
}

int poissonMG_dirichlet(int i, int j, float value) {
   int g;
   if ((i<0)||(j<0)||(i>=grid_list[0]->width)||(j>=grid_list[0]->height)) return POISSON_INDEX_OUT_OF_RANGE;
    switch (solver.type) {
    case TYPE_CARTESIAN :
    for (g=0; g<solver.grid_levels; g++) {
        _dirichlet(grid_list[g],i*grid_list[g]->r,j*grid_list[g]->r,value);   
    }
    break;
    case TYPE_AXIAL :
    for (g=0; g<solver.grid_levels; g++) {
        _dirichlet(grid_list[g],i*grid_list[g]->r,j*grid_list[g]->r,value);   
    }
    break;
    case TYPE_AXIAL_VECT :
    for (g=0; g<solver.grid_levels; g++) {
        _dirichlet(grid_list[g],i*grid_list[g]->r,j*grid_list[g]->r,value);   
    }
    break;
    case TYPE_POLAR :
    for (g=0; g<solver.grid_levels; g++) {
        _dirichlet(grid_list[g],i*grid_list[g]->r,j*grid_list[g]->r,value);   
    }
    break;
    }
    return 0;    
}

int _dirichlet_borders(grid_def* grid, float value) {
    int ig,jg,k;
     for (ig=0; ig<grid->width; ig++) { // dirichlet sur les bords hauts et bas
         k = (ig+1)+grid->full_width;
         grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
         grid->C[k] = 1;
         grid->U[k] = grid->F[k] = value;
         grid->GT[k] = 0;
         k = (ig+1)+(grid->height)*grid->full_width;
         grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
         grid->C[k] = 1;
         grid->U[k] = grid->F[k] = value;
         grid->GT[k] = 0;
     }
     for (jg=0; jg<grid->height; jg++) { // dirichlet sur les bords gauche et droite
         k = 1+(jg+1)*grid->full_width;
         grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
         grid->C[k] = 1;
         grid->U[k] = grid->F[k] = value;
         grid->GT[k] = 0;
         k = grid->width+(jg+1)*grid->full_width;
         grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
         grid->C[k] = 1;
         grid->U[k] = grid->F[k] =value;
         grid->GT[k] = 0;
     }
     return 0;   
}

int _dirichlet_borders_axial(grid_def* grid, float value) {
    int ig,jg,k;
     for (ig=0; ig<grid->width; ig++) { // dirichlet sur le bord haut
         k = (ig+1)+grid->full_width*grid->height;
         grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
         grid->C[k] = 1;
         grid->U[k] = grid->F[k] = value;
         grid->GT[k] = 0;
     }
     for (jg=0; jg<grid->height; jg++) { // dirichlet sur les bords gauche et droite
         k = 1+(jg+1)*grid->full_width;
         grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
         grid->C[k] = 1;
         grid->U[k] = grid->F[k] = value;
         grid->GT[k] = 0;
         k = grid->width+(jg+1)*grid->full_width;
         grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
         grid->C[k] = 1;
         grid->U[k] = grid->F[k] =value;
         grid->GT[k] = 0;
     }
     return 0;   
}

int _dirichlet_borders_polar(grid_def* grid, float value) {
    int ig,k;
     for (ig=0; ig<grid->width; ig++) { // dirichlet sur le bord haut
         k = (ig+1)+grid->full_width*grid->height;
         grid->D[k] = grid->G[k] = grid->H[k] = grid->B[k] = 0;
         grid->C[k] = 1;
         grid->U[k] = grid->F[k] = value;
         grid->GT[k] = 0;
     }
     return 0;   
}

int poissonMG_dirichlet_borders(float value) {
    int g;
    switch (solver.type) {
    case TYPE_CARTESIAN :
    for (g=0; g<solver.grid_levels; g++) {
        _dirichlet_borders(grid_list[g],value);   
    }
    break;
    case TYPE_AXIAL :
    for (g=0; g<solver.grid_levels; g++) {
        _dirichlet_borders_axial(grid_list[g],value);    
    }
    break;
    case TYPE_AXIAL_VECT :
    for (g=0; g<solver.grid_levels; g++) {
        _dirichlet_borders_axial(grid_list[g],value);   
    }
    break;
    case TYPE_POLAR :
    for (g=0; g<solver.grid_levels; g++) {
        _dirichlet_borders_polar(grid_list[g],value);   
    }
    break;
    }
    return 0;
}



int _dirichlet_polygon(grid_def* grid, int i_start, int j_start, intArray li_list, intArray lj_list, float value) {
    int k,nig,njg,ig,jg,ig_start,jg_start,id,jd;
    if ((lj_list.ndims!=1)||(lj_list.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (li_list.dims[0]!=lj_list.dims[0]) return POISSON_BAD_ARRAY_DIM;
    
    ig_start = i_start*grid->r;
    jg_start = j_start*grid->r;
    
    for (k=0; k<li_list.dims[0]; k++) {
        nig = li_list.data[k]*grid->r;
        njg = lj_list.data[k]*grid->r;
        if (nig!=0) {
            
               jd = 0;
               id = 1;
               if (nig<0) id = -1;
               if (njg!=0) return POISSON_BAD_POLYGON_DEFINITION;
               //if (last&&(k==li_list.dims[0]-1)) nig+=id;
               jg = jg_start;
               for (ig=ig_start; ig!=ig_start+nig+id; ig+=id) {
                   grid->P[ig+grid->width*jg] = MASK_DIRICHLET;
                   
                   _dirichlet(grid,ig,jg,value);
               }
               
               //printf("%d : %d,%d\n",grid->g,ig-1,jg);
        }
        if (njg!=0) {
               id = 0;
               jd = 1;
               if (njg<0) jd = -1;
               if (nig!=0) return POISSON_BAD_POLYGON_DEFINITION;
               //if (last&&(k==li_list.dims[0]-1)) njg+=jd;
               ig = ig_start;
               for (jg=jg_start; jg!=jg_start+njg+jd; jg+=jd) {
                   grid->P[ig+grid->width*jg] = MASK_DIRICHLET;
                   //if (jg==jg_start) printf("%d : %d,%d\n",grid->g,ig,jg);
                   _dirichlet(grid,ig,jg,value);
               }
        }
        
        ig_start = ig;
        jg_start = jg;
    }
    return 0;  
}

int poissonMG_dirichlet_polygon(int i_start, int j_start, intArray li_list, intArray lj_list, float value) {
    int g,err;
    switch (solver.type) {
    case TYPE_CARTESIAN :
    for (g=0; g<solver.grid_levels; g++) {
        err = _dirichlet_polygon(grid_list[g],i_start,j_start,li_list,lj_list,value);
        if (err<0) return err;
    }
    break;
    case TYPE_AXIAL :
    for (g=0; g<solver.grid_levels; g++) {
        err = _dirichlet_polygon(grid_list[g],i_start,j_start,li_list,lj_list,value);
        if (err<0) return err;
    }
    break;
    case TYPE_AXIAL_VECT :
    for (g=0; g<solver.grid_levels; g++) {
        err = _dirichlet_polygon(grid_list[g],i_start,j_start,li_list,lj_list,value);
        if (err<0) return err;
    }
    break;
    case TYPE_POLAR :
    for (g=0; g<solver.grid_levels; g++) {
        err = _dirichlet_polygon(grid_list[g],i_start,j_start,li_list,lj_list,value);
        if (err<0) return err;
    }
    break;
    }
    return 0;  
}

int _neumann(grid_def* grid, int igp, int jgp, int ig, int jg, int ign, int jgn, float source, float derivX, float derivY) {
    //igp,jgp : previous point
    //ig,jg : point
    //ign,jgn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx,dy,dx2,dy2,r;
    dx = grid->dx;
    dy = grid->dy;
    dx2=dx*dx;
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = ig-igp;
    di2 = ign-ig;
    dj1 = jg-jgp;
    dj2 = jgn-jg;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            
            grid->D[k] = grid->G[k] = 0.5*r; grid->H[k] = 1.0; grid->B[k] = 0.0; grid->C[k] = -1.0-r; grid->F[k] = dy*derivY+0.5*dy2*source;
        }
        else if (di1==-1) {
            
            grid->D[k] = grid->G[k] = 0.5*r; grid->H[k] = 0.0; grid->B[k] = 1.0; grid->C[k] = -1.0-r; grid->F[k] = -dy*derivY+0.5*dy2*source;   
        }
        else if (dj1==1) {
            
            grid->H[k] = grid->B[k] = 0.5; grid->G[k] = r; grid->D[k] = 0.0; grid->C[k] =-1.0-r; grid->F[k] = -dy2/dx*derivX+0.5*dy2*source;
        }
        else if (dj1==-1) {
            
            grid->H[k] = grid->B[k] = 0.5; grid->G[k] = 0.0; grid->D[k] = r; grid->C[k] =-1.0-r; grid->F[k] = dy2/dx*derivX+0.5*dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            grid->G[k] = 0.5*r; grid->D[k] =0.0; grid->H[k] = 0.5; grid->B[k] = 0.0; grid->C[k] = -0.5*(1.0+r); grid->F[k] = -0.5*dy2/dx*derivX+0.5*dy*derivY+0.25*dy2*source;  
        }
        else if ((di1==-1)&&(dj2==-1)) {
            grid->G[k] = 0.0; grid->D[k] = 0.5*r; grid->H[k] = 0.0; grid->B[k] = 0.5; grid->C[k] = -0.5*(1.0+r); grid->F[k] = 0.5*dy2/dx*derivX-0.5*dy*derivY+0.25*dy2*source;   
        }
        else if ((di2==1)&&(dj1==-1)) {
            grid->G[k] = 0.0; grid->D[k] = 0.5*r; grid->H[k] = 0.5; grid->B[k] = 0.0; grid->C[k] = -0.5*(1.0+r); grid->F[k] = 0.5*dy2/dx*derivX+0.5*dy*derivY+0.25*dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
            grid->G[k] = 0.5*r; grid->D[k] = 0.0; grid->H[k] = 0.0; grid->B[k] = 0.5; grid->C[k] = -0.5*(1.0+r); grid->F[k] = -0.5*dy2/dx*derivX-0.5*dy*derivY+0.25*dy2*source;  
        }
        else if ((di1==1)&&(dj2==-1)) {
            grid->G[k] = 0.5*r; grid->D[k] = r; grid->H[k] = 1.0; grid->B[k] = 0.5; grid->C[k] = -1.5*(1.0+r); grid->F[k] = 0.5*dy2/dx*derivX+0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((di1==-1)&&(dj2==1)) {
            grid->G[k] = r; grid->D[k] = 0.5*r; grid->H[k] = 0.5; grid->B[k] = 1.0; grid->C[k] = -1.5*(1.0+r); grid->F[k] = -0.5*dy2/dx*derivX-0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((dj1==-1)&&(di2==-1)) {
            grid->G[k] = 0.5*r; grid->D[k] = r; grid->H[k] = 0.5; grid->B[k] = 1.0; grid->C[k] = -1.5*(1.0+r); grid->F[k] = 0.5*dy2/dx*derivX-0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((dj1==1)&&(di2==1)) {
            grid->G[k] = r; grid->D[k] = 0.5*r; grid->H[k] = 1.0; grid->B[k] = 0.5; grid->C[k] = -1.5*(1.0+r); grid->F[k] = -0.5*dy2/dx*derivX+0.5*dy*derivY+0.75*dy2*source;   
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;
}

int _neumann_axial(grid_def* grid, int igp, int jgp, int ig, int jg, int ign, int jgn, float source, float derivX, float derivY) {
    //igp,jgp : previous point
    //ig,jg : point
    //ign,jgn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx,dy,dx2,dy2,r;
    float hi;
    hi = 1.0/grid->height;
    dx = grid->dx;
    dy = grid->dy;
    dx2=dx*dx;
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = ig-igp;
    di2 = ign-ig;
    dj1 = jg-jgp;
    dj2 = jgn-jg;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            grid->G[k]=grid->D[k]=0.5*(jg+0.25)*hi*r; grid->B[k] = 0.0; grid->H[k] = (jg+0.5)*hi; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (jg*dy*derivY+0.5*(jg+0.25)*dy2*source)*hi;
            
        }
        else if (di1==-1) {
            grid->G[k]=grid->D[k]=0.5*(jg-0.25)*hi*r; grid->B[k] = (jg-0.5)*hi; grid->H[k] = 0.0; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (-jg*dy*derivY+0.5*(jg-0.25)*dy2*source)*hi;
            
        }
        else if (dj1==1) {
            grid->G[k] = jg*hi*r; grid->D[k] = 0.0; grid->H[k] = 0.5*(jg+0.5)*hi; grid->B[k] = 0.5*(jg-0.5)*hi; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (-jg*dy2/dx*derivX+0.5*jg*dy2*source)*hi;
        }
        else if (dj1==-1) {
            grid->G[k] = 0.0; grid->D[k] = jg*hi*r; grid->H[k] = 0.5*(jg+0.5)*hi; grid->B[k] = 0.5*(jg-0.5)*hi; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (jg*dy2/dx*derivX+0.5*jg*dy2*source)*hi;
                
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
              grid->G[k] = 0.5*(jg+0.25)*hi*r; grid->D[k] = 0.0; grid->H[k] = 0.5*(jg+0.5)*hi; grid->B[k] = 0.0; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (-0.5*(jg+0.25)*dy2/dx*derivX+0.5*jg*dy*derivY+0.25*(jg+0.25)*dy2*source)*hi;
              
        }
        else if ((di1==-1)&&(dj2==-1)) {
              grid->G[k] = 0.0; grid->D[k] = 0.5*(jg-0.25)*hi*r; grid->H[k] = 0.0; grid->B[k] = 0.5*(jg-0.5)*hi; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (0.5*(jg-0.25)*dy2/dx*derivX-0.5*jg*dy*derivY+0.25*(jg-0.25)*dy2*source)*hi;
              
        }
        else if ((di2==1)&&(dj1==-1)) {
               grid->G[0] = 0.0; grid->D[k] = 0.5*(jg+0.25)*hi*r; grid->H[k] = 0.5*(jg+0.5)*hi; grid->B[k] = 0.0;  grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (0.5*(jg+0.25)*dy2/dx*derivX-0.5*jg*dy*derivY+0.25*(jg+0.25)*dy2*source)*hi;
        }
        else if ((di2==-1)&&(dj1==1)) {
               grid->G[k] = 0.5*(jg-0.25)*hi*r; grid->D[k] = 0.0; grid->H[k] = 0.0; grid->B[k] = 0.5*(jg-0.5)*hi; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (-0.5*(jg-0.25)*dy2/dx*derivX-0.5*jg*dy*derivY+0.25*(jg-0.25)*dy2*source)*hi;
        }
        else if ((di1==1)&&(dj2==-1)) {
                grid->G[k] = 0.5*(jg+0.25)*hi*r; grid->D[k] = jg*r; grid->H[k] = (jg+0.5)*hi; grid->B[k] = 0.5*(jg-0.5)*hi; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (-0.5*(jg+0.25)*dy2/dx*derivX-0.5*jg*dy*derivY+(0.5*(jg-0.25)+0.25*(jg+0.25))*dy2*source)*hi;
        }
        else if ((di1==-1)&&(dj2==1)) {
                grid->G[k] = jg*hi*r; grid->D[k] = 0.5*(jg-0.25)*hi*r; grid->H[k] = 0.5*(jg+0.5)*hi; grid->B[k] = (jg-0.5)*hi; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = (0.5*(jg-0.25)*dy2/dx*derivX+0.5*jg*dy*derivY+(0.5*(jg+0.25)+0.25*(jg-0.25))*dy2*source)*hi;
        }
        else if ((dj1==-1)&&(di2==-1)) {
                grid->G[k] = 0.5*(jg-0.25)*hi*r; grid->D[k] = jg*r; grid->H[k] = 0.5*(jg+0.5)*hi; grid->B[k] = (jg-0.5)*hi; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] =  (0.5*(jg+0.25)*dy2/dx*derivX-0.5*jg*dy*derivY+(0.5*(jg-0.25)+0.25*(jg+0.25))*dy2*source)*hi;
        }
        else if ((dj1==1)&&(di2==1)) {
               grid->G[k] = jg*r; grid->D[k] = 0.5*(jg+0.25)*hi*r; grid->H[k] = (jg+0.5)*hi; grid->B[k] = 0.5*(jg-0.5)*hi; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] =  (-0.5*(jg-0.25)*dy2/dx*derivX+0.5*jg*dy*derivY+(0.5*(jg+0.25)+0.25*(jg-0.25))*dy2*source)*hi;
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;
}

int _neumann_axial_vect(grid_def *grid,int igp, int jgp, int ig, int jg, int ign, int jgn, float source, float derivX, float derivY) {
    //igp,jgp : previous point
    //ig,jg : point
    //ign,jgn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r,dx,dy;
    dx = grid->dx;
    dy = grid->dy;
    dx2=dx*dx;
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = ig-igp;
    di2 = ign-ig;
    dj1 = jg-jgp;
    dj2 = jgn-jg;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            
            grid->D[k] = grid->G[k] = 0.5*r; grid->H[k] = (jg+1.0)/(jg+0.5); grid->B[k] = 0.0; grid->C[k] = -(jg/(jg+0.5)+r); grid->F[k] = dy*derivY+0.5*dy2*source;
        }
        else if (di1==-1) {
            
            grid->D[k] = grid->G[k] = 0.5*r; grid->H[k] = 0.0; grid->B[k] = (jg-1.0)/(jg-0.5); grid->C[k] = -(jg/(jg-0.5)+r); grid->F[k] = -dy*derivY+0.5*dy2*source;   
        }
        else if (dj1==1) {
            
            grid->G[k] = r; grid->D[k] = 0.0; grid->H[k] = 0.5*(jg+1)/(jg+0.5); grid->B[k] = 0.5*(jg-1)/(jg-0.5); grid->C[k] =-(0.5*jg/(jg+0.5)+0.5*jg/(jg-0.5)+r); grid->F[k] = -dy2/dx*derivX+0.5*dy2*source;
        }
        else if (dj1==-1) {
            
            grid->G[k] = 0.0; grid->D[k] = r; grid->H[k] = 0.5*(jg+1)/(jg+0.5); grid->B[k] = 0.5*(jg-1)/(jg-0.5); grid->C[k] =-(0.5*jg/(jg+0.5)+0.5*jg/(jg-0.5)+r); grid->F[k] = dy2/dx*derivX+0.5*dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            grid->G[k] = 0.5*r; grid->D[k] =0.0; grid->H[k] = 0.5*(jg+1.0)/(jg+0.5); grid->B[k] = 0.0; grid->C[k] = -(0.5*jg/(jg+0.5)+r); grid->F[k] = -0.5*dy2/dx*derivX+0.5*dy*derivY+0.25*dy2*source;  
        }
        else if ((di1==-1)&&(dj2==-1)) {
            grid->G[k] = 0.0; grid->D[k] = 0.5*r; grid->H[k] = 0.0; grid->B[k] = 0.5*(jg-1.0)/(jg-0.5); grid->C[k] = -(0.5*jg/(jg-0.5)+r); grid->F[k] = 0.5*dy2/dx*derivX-0.5*dy*derivY+0.25*dy2*source;   
        }
        else if ((di2==1)&&(dj1==-1)) {
            grid->G[k] = 0.0; grid->D[k] = 0.5*r; grid->H[k] = 0.5*(jg+1.0)/(jg+0.5); grid->B[k] = 0.0; grid->C[k] = -(0.5*jg/(jg+0.5)+r); grid->F[k] = 0.5*dy2/dx*derivX+0.5*dy*derivY+0.25*dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
            grid->G[k] = 0.5*r; grid->D[k] = 0.0; grid->H[k] = 0.0; grid->B[k] = 0.5*(jg-1.0)/(jg-0.5); grid->C[k] = -(0.5*jg/(jg-0.5)+r); grid->F[k] = -0.5*dy2/dx*derivX-0.5*dy*derivY+0.25*dy2*source;  
        }
        else if ((di1==1)&&(dj2==-1)) {
            grid->G[k] = 0.5*r; grid->D[k] = r; grid->H[k] = (jg+1.0)/(jg+0.5); grid->B[k] = 0.5*(jg-1.0)/(jg-0.5); grid->C[k] = -(jg/(jg+0.5)+0.5*jg/(jg-0.5)+1.5*r); grid->F[k] = 0.5*dy2/dx*derivX+0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((di1==-1)&&(dj2==1)) {
            grid->G[k] = r; grid->D[k] = 0.5*r; grid->H[k] = 0.5*(jg+1.0)/(jg+0.5); grid->B[k] = (jg-1.0)/(jg-0.5); grid->C[k] = -(0.5*jg/(jg+0.5)+jg/(jg-0.5)+1.5*r); grid->F[k] = -0.5*dy2/dx*derivX-0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((dj1==-1)&&(di2==-1)) {
            grid->G[k] = 0.5*r; grid->D[k] = r; grid->H[k] = 0.5*(jg+1.0)/(jg+0.5); grid->B[k] = (jg-1.0)/(jg-0.5); grid->C[k] = -(0.5*jg/(jg+0.5)+jg/(jg-0.5)+1.5*r); grid->F[k] = 0.5*dy2/dx*derivX-0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((dj1==1)&&(di2==1)) {
            grid->G[k] = r; grid->D[k] = 0.5*r; grid->H[k] = (jg+1.0)/(jg+0.5); grid->B[k] = 0.5*(jg-1.0)/(jg-0.5); grid->C[k] = -(jg/(jg+0.5)+0.5*jg/(jg-0.5)+1.5*r); grid->F[k] = -0.5*dy2/dx*derivX+0.5*dy*derivY+0.75*dy2*source;   
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;
}


int _neumann_polar(grid_def* grid, int igp, int jgp, int ig, int jg, int ign, int jgn, float source, float derivX, float derivY) {
    //igp,jgp : previous point
    //ig,jg : point
    //ign,jgn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dr,dtheta,dr2,dtheta2,h,b;
    float hi,hi2;
    hi = 1.0/(grid->height);
    hi = 1.0;
    hi2 = hi*hi;
    dtheta = grid->dx;
    dr = grid->dy;
    dr2=dr*dr;
    dtheta2=dtheta*dtheta;
    di1 = ig-igp;
    di2 = ign-ig;
    dj1 = jg-jgp;
    dj2 = jgn-jg;
    h = (jg+0.5)*jg*dtheta2;
    b = (jg-0.5)*jg*dtheta2;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            grid->G[k]=grid->D[k]=0.5; grid->B[k] = 0.0; grid->H[k] = h*0.5; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*0.5*(jg+0.25)*jg*hi2*dtheta2*dr2+derivY*jg*jg*hi2*dr*dtheta2;
            
        }
        else if (di1==-1) {
            grid->G[k]=grid->D[k]=0.5; grid->B[k] = b*0.5; grid->H[k] = 0.0; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*0.5*(jg-0.25)*jg*hi2*dtheta2*dr2-derivY*jg*jg*hi2*dr*dtheta2;
            
        }
        else if (dj1==1) {
            grid->G[k] = 1.0; grid->D[k] = 0.0; grid->H[k] = h*0.5; grid->B[k] = b*0.5; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*0.5*jg*jg*hi2*dtheta2*dr2-derivX*jg*hi*dr*dtheta;
        }
        else if (dj1==-1) {
            grid->G[k] = 0.0; grid->D[k] = 1.0; grid->H[k] = h*0.5; grid->B[k] = b*0.5; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*0.5*jg*jg*hi2*dtheta2*dr2+derivX*jg*hi*dr*dtheta;
                
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
              grid->G[k] = 0.5; grid->D[k] = 0.0; grid->H[k] = h*0.5; grid->B[k] = 0.0; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*(0.25*jg+0.0625)*jg*hi2*dr2*dtheta2-derivX*0.5*jg*hi*dr*dtheta2+derivY*0.5*jg*jg*hi2*dr*dtheta2;
              
        }
        else if ((di1==-1)&&(dj2==-1)) {
              grid->G[k] = 0.0; grid->D[k] = 0.5; grid->H[k] = 0.0; grid->B[k] = b*0.5; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*(0.25*jg-0.0625)*jg*hi2*dr2*dtheta2+derivX*0.5*jg*hi*dr*dtheta2-derivY*0.5*jg*jg*hi2*dr*dtheta2;
              
        }
        else if ((di2==1)&&(dj1==-1)) {
               grid->G[0] = 0.0; grid->D[k] = 0.5; grid->H[k] = 0.5*h; grid->B[k] = 0.0;  grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*(0.25*jg+0.0625)*jg*hi2*dr2*dtheta2+derivX*0.5*jg*hi*dr*dtheta+derivY*0.5*jg*jg*hi2*dr*dtheta2;
        }
        else if ((di2==-1)&&(dj1==1)) {
               grid->G[k] = 0.5; grid->D[k] = 0.0; grid->H[k] = 0.0; grid->B[k] = b*0.5; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*(0.25*jg-0.0625)*jg*hi2*dr2*dtheta2-derivX*0.5*jg*hi*dr*dtheta-derivY*0.5*jg*jg*hi2*dr*dtheta2;
        }
        else if ((di1==1)&&(dj2==-1)) {
                grid->G[k] = 0.5; grid->D[k] = 1.0; grid->H[k] = h; grid->B[k] = b*0.5; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*(0.75*jg+0.0625)*jg*hi2*dr2*dtheta2+derivX*0.5*jg*hi*dr*dtheta+derivY*0.5*jg*jg*hi2*dr*dtheta2 ;
        }
        else if ((di1==-1)&&(dj2==1)) {
                grid->G[k] = 1; grid->D[k] = 0.5; grid->H[k] = h*0.5; grid->B[k] = b; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*(0.75*jg-0.0625)*jg*hi2*dr2*dtheta2-derivX*0.5*jg*hi*dr*dtheta-derivY*0.5*jg*jg*hi2*dr*dtheta2;
        }
        else if ((dj1==-1)&&(di2==-1)) {
                grid->G[k] = 0.5; grid->D[k] = 1; grid->H[k] = h*0.5; grid->B[k] = b; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = source*(0.75*jg-0.0625)*jg*hi2*dr2*dtheta2+derivX*0.5*jg*hi*dr*dtheta-derivY*0.5*jg*jg*hi2*dr*dtheta2 ;
        }
        else if ((dj1==1)&&(di2==1)) {
               grid->G[k] = 1; grid->D[k] = 0.5; grid->H[k] = h; grid->B[k] = b*0.5; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] =  source*(0.75*jg+0.0625)*jg*hi2*dr2*dtheta2-derivX*0.5*jg*hi*dr*dtheta+derivY*0.5*jg*jg*hi2*dr*dtheta2;
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;
}

int _neumann_borders(grid_def* grid, float source, float derivX1, float derivX2, float derivY1, float derivY2) {
    int ig,jg;
    int err;
    err=_neumann(grid,0,1,0,0,1,0,source,derivX1,derivY1); // coin inférieur gauche
    if (err!=0) return err;
    for (ig=1; ig<grid->width-1; ig++) { // bord inférieur
        err=_neumann(grid,ig-1,0,ig,0,ig+1,0,source,0,derivY1);
        if (err!=0) return err;
    }
    err=_neumann(grid,ig-1,0,ig,0,ig,1,source,derivX2,derivY1); // coin inférieur droit
    if (err!=0) return err;
    for (jg=1; jg<grid->height-1; jg++) { // bord droit
        err=_neumann(grid,ig,jg-1,ig,jg,ig,jg+1,source,derivX2,0);
        if (err!=0) return err;
    }
    err=_neumann(grid,ig,jg-1,ig,jg,ig-1,jg,source,derivX2,derivY2); // coin supérieur droit
    if (err!=0) return err;
    for (ig=grid->width-2; ig>0; ig--) { // bord supérieur
        err=_neumann(grid,ig+1,jg,ig,jg,ig-1,jg,source,0,derivY2);
        if (err!=0) return err;
    }
    err=_neumann(grid,ig+1,jg,ig,jg,ig,jg-1,source,derivX1,derivY2); // coin supérieur gauche
    if (err!=0) return err;
    for (jg=grid->height-2; jg>0; jg--) { // bord gauche
        err=_neumann(grid,ig,jg+1,ig,jg,ig,jg-1,source,derivX1,0);
        if (err!=0) return err;
    }
    return 0;  
}

int _neumann_borders_axial(grid_def* grid, float source, float derivX1, float derivX2, float derivY1, float derivY2) {
    int ig,jg,k;
    int err;
    float dx2,dy2,r;
    dx2=grid->dx*grid->dx;
    dy2=grid->dy*grid->dy;
    r=dy2/dx2;
    // coin inférieur gauche
    ig = 0; jg = 0;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    grid->G[k] = 0.0; grid->D[k] = r/8.0; grid->H[k] = 0.25; grid->B[k] = 0.0; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2/grid->dx/8.0*derivX1+dy2/16.0*source;
    // coin inférieur droit
    ig = grid->width-1;
    k = (ig+1)+grid->full_width*(jg+1);
    grid->G[k] = r/8.0; grid->D[k] = 0.0; grid->H[k] = 0.25; grid->B[k] = 0.0; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = -dy2/grid->dx/8.0*derivX2+dy2/16.0*source;
    for (jg=1; jg<grid->height-1; jg++) { // bord droit
        err=_neumann_axial(grid,ig,jg-1,ig,jg,ig,jg+1,source,derivX2,0);
        if (err!=0) return err;
    }
    err=_neumann_axial(grid,ig,jg-1,ig,jg,ig-1,jg,source,derivX2,derivY2); // coin supérieur droit
    if (err!=0) return err;
    for (ig=grid->width-2; ig>0; ig--) { // bord supérieur
        err=_neumann_axial(grid,ig+1,jg,ig,jg,ig-1,jg,source,0,derivY2);
        if (err!=0) return err;
    }
    err=_neumann_axial(grid,ig+1,jg,ig,jg,ig,jg-1,source,derivX1,derivY2); // coin supérieur gauche
    if (err!=0) return err;
    for (jg=grid->height-2; jg>0; jg--) { // bord gauche
        err=_neumann_axial(grid,ig,jg+1,ig,jg,ig,jg-1,source,derivX1,0);
        if (err!=0) return err;
    }
    return 0;   
}

int _neumann_borders_axial_vect(grid_def* grid, float source, float derivX1, float derivX2, float derivY1, float derivY2) {
    int ig,jg;
    int err;
    ig = grid->width-1;
    for (jg=1; jg<grid->height-1; jg++) { // bord droit
        err=_neumann_axial_vect(grid,ig,jg-1,ig,jg,ig,jg+1,source,derivX2,0);
        if (err!=0) return err;
    }
    err=_neumann_axial_vect(grid,ig,jg-1,ig,jg,ig-1,jg,source,derivX2,derivY2); // coin supérieur droit
    if (err!=0) return err;
    for (ig=grid->width-2; ig>0; ig--) { // bord supérieur
        err=_neumann_axial_vect(grid,ig+1,jg,ig,jg,ig-1,jg,source,0,derivY2);
        if (err!=0) return err;
    }
    err=_neumann_axial_vect(grid,ig+1,jg,ig,jg,ig,jg-1,source,derivX1,derivY2); // coin supérieur gauche
    if (err!=0) return err;
    for (jg=grid->height-2; jg>0; jg--) { // bord gauche
        err=_neumann_axial_vect(grid,ig,jg+1,ig,jg,ig,jg-1,source,derivX1,0);
        if (err!=0) return err;
    }
    return 0;
}


int _neumann_border_ext_polar(grid_def* grid, float source, float derivR){
    int ig,jg,k;
    float dtheta2,dr2;
    dtheta2 = grid->dx*grid->dx;
    dr2 = grid->dy*grid->dy;
    jg = grid->height-1;
    for (ig=0; ig<grid->width; ig++) {
           k = (ig+1)+grid->full_width*(jg+1);
           grid->D[k] = grid->G[k] = 1.0;
           grid->H[k] = 0.0;
           grid->B[k] = (2*jg-1)*jg*dtheta2;
           grid->C[k] = -grid->B[k]-2.0;
           grid->F[k] = source*(jg-0.25)*jg*dtheta2*dr2-derivR*2*jg*jg*grid->dy*dtheta2;
    }
    return 0;
}

int _neumann_border_int_polar(grid_def* grid, int j, float source, float derivR){
    int ig,jg,k;
    float dtheta2,dr2;
    dtheta2 = grid->dx*grid->dx;
    dr2 = grid->dy*grid->dy;
    jg = j*grid->r;
    grid->jg_min = jg;
    for (ig=0; ig<grid->width; ig++) {
           k = (ig+1)+grid->full_width*(jg+1);
           grid->D[k] = grid->G[k] = 1.0;
           grid->B[k] = 0.0;
           grid->H[k] = (2*jg+1)*jg*dtheta2;
           grid->C[k] = -grid->H[k]-2.0;
           grid->F[k] = source*(jg+0.25)*jg*dtheta2*dr2+derivR*2*jg*jg*grid->dy*dtheta2;
    }
    return 0;
}



int poissonMG_neumann_border_ext_polar(float source, float derivR){
    int g;
    for (g=0; g<solver.grid_levels; g++) {
        _neumann_border_ext_polar(grid_list[g],source,derivR);
    }
    return 0;
}

int poissonMG_neumann_border_int_polar(int j, float source, float derivR){
    int g;
    for (g=0; g<solver.grid_levels; g++) {
        _neumann_border_int_polar(grid_list[g],j,source,derivR);
    }
    return 0;
}

int poissonMG_neumann_borders(float source, float derivX1, float derivX2, float derivY1, float derivY2 ) {
    int g;
    switch (solver.type) {
    case TYPE_CARTESIAN :
    for (g=0; g<solver.grid_levels; g++) {
        _neumann_borders(grid_list[g],source,derivX1,derivX2,derivY1,derivY2);  
    }
    break;
    case TYPE_AXIAL:
    for (g=0; g<solver.grid_levels; g++) {
        _neumann_borders_axial(grid_list[g],source,derivX1,derivX2,derivY1,derivY2);  
    }
    break;
    case TYPE_AXIAL_VECT:
    for (g=0; g<solver.grid_levels; g++) {
        _neumann_borders_axial_vect(grid_list[g],source,derivX1,derivX2,derivY1,derivY2);  
    }
    break;
    }
    return 0;
}

int _neumann_polygon(grid_def* grid, int i_start, int j_start, intArray li_list, intArray lj_list, floatArray derivX_list, floatArray derivY_list,float source) {
    int k,nig,njg,ig,jg,ig_p,jg_p,id,jd;
    int ig_start, jg_start, ig_first, jg_first, ign_first, jgn_first;
    float derivX,derivY;
    derivX = derivY = 0;
    ig_p = jg_p = ig_first = jg_first = ign_first = jgn_first = 0;
    if ((lj_list.ndims!=1)||(lj_list.ndims!=1)||(derivX_list.ndims!=1)||(derivY_list.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if ((li_list.dims[0]!=lj_list.dims[0])||(derivX_list.dims[0]!=derivY_list.dims[0])||(li_list.dims[0]!=derivX_list.dims[0])) return POISSON_BAD_ARRAY_DIM;
    ig_start = i_start*grid->r;
    jg_start = j_start*grid->r;
    for (k=0; k<li_list.dims[0]; k++) {
        nig = li_list.data[k]*grid->r;
        njg = lj_list.data[k]*grid->r;
        if (nig!=0) {
               if (njg!=0) return POISSON_BAD_POLYGON_DEFINITION;
               jd = 0;
               id = 1;
               if (nig<0) id = -1;
               derivY = derivY_list.data[k];
               if (k==0) { // premier segment
                   ig_first = ig_p = ig_start;
                   jg_first = jg_p = jg_start;
                   ig_start+=id;
                   nig--;
                   ign_first = ig_start;
                   jgn_first = jg_start;
                   derivX = derivX_list.data[derivX_list.dims[0]-1];
               }
               else if (k==li_list.dims[0]-1) { //dernier segment
                   derivX = derivX_list.data[0];
               }
               else {
                    derivX = derivX_list.data[k+1];   
               }
               jg = jg_start;
               for (ig=ig_start; ig!=ig_start+nig; ig+=id) {
                   grid->P[ig+grid->width*jg] = MASK_NEUMANN;
                   solver.neumann_function(grid,ig_p, jg_p, ig, jg, ig+id, jg, derivX,derivY,source);
                   ig_p = ig; jg_p = jg;
                   
               } 
        }
        if (njg!=0) {
               if (nig!=0) return POISSON_BAD_POLYGON_DEFINITION;
               id = 0;
               jd = 1;
               if (njg<0) jd = -1;
               derivX = derivX_list.data[k];
               if (k==0) {
                   ig_first = ig_p = ig_start;
                   jg_first = jg_p = jg_start;
                   jg_start+=jd;
                   njg--;
                   ign_first = ig_start;
                   jgn_first = jg_start;
                   derivY = derivY_list.data[derivY_list.dims[0]-1];
               }
               else if (k==li_list.dims[0]-1) {
                    derivY = derivY_list.data[0];   
               }
               else {
                    derivY = derivY_list.data[k+1];   
               }
               ig = ig_start;
               for (jg=jg_start; jg!=jg_start+njg; jg+=jd) {
                   grid->P[ig+grid->width*jg] = MASK_NEUMANN;
                   solver.neumann_function(grid,ig_p, jg_p, ig, jg, ig, jg+jd, derivX,derivY,source);
                   ig_p = ig; jg_p = jg;
                  
               }
               
        }
        ig_start = ig;
        jg_start = jg;
        
    }
    grid->P[ig_first+grid->width*jg_first] = MASK_NEUMANN;
    solver.neumann_function(grid, ig_p, jg_p, ig_first, jg_first, ign_first, jgn_first, derivX,derivY,source);
    return 0;  
}



int poissonMG_neumann_polygon(int i_start, int j_start, intArray li_list, intArray lj_list, floatArray derivX_list, floatArray derivY_list, float source) {
    int g,err;
    for (g=0; g<solver.grid_levels; g++) {
        err = _neumann_polygon(grid_list[g],i_start,j_start,li_list,lj_list,derivX_list,derivY_list,source);
        if (err<0) return err;
    }
    return 0;    
}

 // interface entre deux milieux différents
int _interface(grid_def* grid,int igp, int jgp, int ig, int jg, int ign, int jgn, float source,float a1, float a2) {
    //igp,jgp : previous point
    //ig,jg : point
    //ign,jgn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r;
    dx2=grid->dx*grid->dx; 
    dy2=grid->dy*grid->dy;
    r=dy2/dx2;
    di1 = ig-igp;
    di2 = ign-ig;
    dj1 = jg-jgp;
    dj2 = jgn-jg;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            grid->D[k] = grid->G[k] =r*(a1+a2)/2; grid->H[k] = a1; grid->B[k] = a2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source;
        }
        else if (di1==-1) {
            grid->D[k] = grid->G[k] = r*(a1+a2)/2; grid->H[k] = a2; grid->B[k] = a1; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source;   
        }
        else if (dj1==1) {
            grid->H[k] = grid->B[k] = (a1+a2)/2; grid->G[k] = a1*r; grid->D[k] = a2*r; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source; 
        }
        else if (dj1==-1) {
            grid->H[k] = grid->B[k] = (a1+a2)/2; grid->G[k] = a2*r; grid->D[k] = a1*r; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            grid->G[k] = r*(a1+a2)/2; grid->D[k] = r*a2; grid->B[k] = a2; grid->H[k] = (a1+a2)/2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source; 
        }
        else if ((di1==-1)&&(dj2==-1)) {
            grid->G[k] = r*a2; grid->D[k] = r*(a1+a2)/2; grid->B[k] = (a1+a2)/2; grid->H[k] = a2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source;  
        }
        else if ((di2==1)&&(dj1==-1)) {
            grid->G[k] = r*a2; grid->D[k] = r*(a1+a2)/2; grid->B[k] = a2; grid->H[k] = (a1+a2)/2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
             grid->G[k] = r*(a1+a2)/2; grid->D[k] = r*a2; grid->B[k] = (a1+a2)/2; grid->H[k] = a2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source; 
        }
        else if ((di1==1)&&(dj2==-1)) {
           grid->G[k] = r*(a1+a2)/2; grid->D[k] = r*a1; grid->B[k] = (a1+a2)/2; grid->H[k] = a1; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source; 
        }
        else if ((di1==-1)&&(dj2==1)) {
           grid->G[k] = r*a1; grid->D[k] = r*(a1+a2)/2; grid->B[k] = a1; grid->H[k] = (a1+a2)/2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source; 
        }
        else if ((dj1==-1)&&(di2==-1)) {
            grid->G[k] = r*(a1+a2)/2; grid->D[k] = r*a1; grid->B[k] = a1; grid->H[k] = (a1+a2)/2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source; 
        }
        else if ((dj1==1)&&(di2==1)) {
             grid->G[k] = r*a1; grid->D[k] = r*(a1+a2)/2; grid->B[k] = (a1+a2)/2; grid->H[k] = a1; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = dy2*source;  
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;       
}

int _interface_axial(grid_def *grid,int igp, int jgp, int ig, int jg, int ign, int jgn, float source,float a1, float a2) { // interface entre deux milieux différents
    //igp,jgp : previous point
    //ig,jg : point
    //ign,jgn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r;
    float dx,dy;
    float hi;
    hi = grid->height;
    dx = grid->dx;
    dy = grid->dy;
    dx2=dx*dx;
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = ig-igp;
    di2 = ign-ig;
    dj1 = jg-jgp;
    dj2 = jgn-jg;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            grid->D[k] = grid->G[k] =jg*hi*r*(a1+a2)/2; grid->H[k] = (jg+0.5)*hi*a1; grid->B[k] = (jg-0.5)*hi*a2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source;
        }
        else if (di1==-1) {
            grid->D[k] = grid->G[k] = jg*hi*r*(a1+a2)/2; grid->H[k] = (jg+0.5)*hi*a2; grid->B[k] = (jg-0.5)*hi*a1; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source;   
        }
        else if (dj1==1) {
            grid->H[k] = (jg+0.5)*hi*(a1+a2)/2; grid->B[k] = (jg-0.5)*hi*(a1+a2)/2; grid->G[k] = jg*hi*a1*r; grid->D[k] = jg*hi*a2*r; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source; 
        }
        else if (dj1==-1) {
            grid->H[k] = (jg+0.5)*hi*(a1+a2)/2;  grid->B[k] =(jg-0.5)*hi*(a1+a2)/2; grid->G[k] = jg*hi*a2*r; grid->D[k] = jg*hi*a1*r; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            grid->G[k] = jg*hi*r*(a1+a2)/2; grid->D[k] = jg*hi*r*a2; grid->B[k] = (jg*-0.5)*hi*a2; grid->H[k] = (jg*+0.5)*hi*(a1+a2)/2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source; 
        }
        else if ((di1==-1)&&(dj2==-1)) {
            grid->G[k] = jg*hi*r*a2; grid->D[k] = jg*hi*r*(a1+a2)/2; grid->B[k] = (jg*-0.5)*hi*(a1+a2)/2; grid->H[k] = (jg*+0.5)*hi*a2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source;  
        }
        else if ((di2==1)&&(dj1==-1)) {
            grid->G[k] = jg*hi*r*a2; grid->D[k] = jg*hi*r*(a1+a2)/2; grid->B[k] = (jg*-0.5)*hi*a2; grid->H[k] = (jg*+0.5)*hi*(a1+a2)/2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
             grid->G[k] = jg*hi*r*(a1+a2)/2; grid->D[k] = jg*hi*r*a2; grid->B[k] = (jg*-0.5)*hi*(a1+a2)/2; grid->H[k] = (jg*+0.5)*hi*a2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source; 
        }
        else if ((di1==1)&&(dj2==-1)) {
           grid->G[k] = jg*hi*r*(a1+a2)/2; grid->D[k] = jg*hi*r*a1; grid->B[k] = (jg*-0.5)*hi*(a1+a2)/2; grid->H[k] = (jg*+0.5)*hi*a1; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source; 
        }
        else if ((di1==-1)&&(dj2==1)) {
           grid->G[k] = jg*hi*r*a1; grid->D[k] = jg*hi*r*(a1+a2)/2; grid->B[k] = (jg*-0.5)*hi*a1; grid->H[k] = (jg*+0.5)*hi*(a1+a2)/2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source; 
        }
        else if ((dj1==-1)&&(di2==-1)) {
            grid->G[k] = jg*hi*r*(a1+a2)/2; grid->D[k] = jg*hi*r*a1; grid->B[k] = (jg*-0.5)*hi*a1; grid->H[k] = (jg*+0.5)*hi*(a1+a2)/2; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source; 
        }
        else if ((dj1==1)&&(di2==1)) {
             grid->G[k] = jg*hi*r*a1; grid->D[k] = jg*hi*r*(a1+a2)/2; grid->B[k] = (jg*-0.5)*hi*(a1+a2)/2; grid->H[k] = (jg*+0.5)*hi*a1; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = jg*hi*dy2*source;  
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;       
}

int _interface_axial_vect(grid_def* grid, int igp, int jgp, int ig, int jg, int ign, int jgn, float source,float a1, float a2) { // interface entre deux milieux différents
    //igp,jgp : previous point
    //ig,jg : point
    //ign,jgn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r;
    float h,b,h0,b0;
    float dx,dy;
    dx = grid->dx;
    dy = grid->dy;
    dx2=dx*dx; 
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = ig-igp;
    di2 = ign-ig;
    dj1 = jg-jgp;
    dj2 = jgn-jg;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    h = (jg+1.0)/(jg+0.5);
    b = (jg-1.0)/(jg-0.5);
    h0 = jg/(jg+0.5);
    b0 = jg/(jg-0.5);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            grid->D[k] = grid->G[k] =r*(a1+a2)/2; grid->H[k] = a1*h; grid->B[k] = a2*b; grid->C[k] = -(grid->G[k]+grid->D[k]+a1*h0+a2*b0); grid->F[k] = dy2*source;
        }
        else if (di1==-1) {
            grid->D[k] = grid->G[k] = r*(a1+a2)/2; grid->H[k] = a2*h; grid->B[k] = a1*b; grid->C[k] = -(grid->G[k]+grid->D[k]+a2*h0+a1*b0); grid->F[k] = dy2*source;   
        }
        else if (dj1==1) {
            grid->G[k] = a1*r; grid->D[k] = a2*r; grid->H[k] = (a1+a2)/2*h; grid->B[k] = (a1+a2)/2*b; grid->C[k] = -(grid->G[k]+grid->D[k]+(a1+a2)/2*(h0+b0)); grid->F[k] = dy2*source; 
        }
        else if (dj1==-1) {
            grid->G[k] = a2*r; grid->D[k] = a1*r; grid->H[k] = (a1+a2)/2*h; grid->B[k] = (a1+a2)/2*b; grid->C[k] = -(grid->G[k]+grid->D[k]+(a1+a2)/2*(h0+b0)); grid->F[k] = dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            grid->G[k] = r*(a1+a2)/2; grid->D[k] = r*a2; grid->B[k] = a2*b; grid->H[k] = (a1+a2)/2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+(a1+a2)/2*h0+a2*b0); grid->F[k] = dy2*source; 
        }
        else if ((di1==-1)&&(dj2==-1)) {
            grid->G[k] = r*a2; grid->D[k] = r*(a1+a2)/2; grid->B[k] = (a1+a2)/2*b; grid->H[k] = a2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+a2*h0+(a1+a2)/2*b0); grid->F[k] = dy2*source;  
        }
        else if ((di2==1)&&(dj1==-1)) {
            grid->G[k] = r*a2; grid->D[k] = r*(a1+a2)/2; grid->B[k] = a2*b; grid->H[k] = (a1+a2)/2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+(a1+a2)/2*h0+a2*b0); grid->F[k] = dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
             grid->G[k] = r*(a1+a2)/2; grid->D[k] = r*a2; grid->B[k] = (a1+a2)/2*b; grid->H[k] = a2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+a2*h0+(a1+a2)/2*b0); grid->F[k] = dy2*source; 
        }
        else if ((di1==1)&&(dj2==-1)) {
           grid->G[k] = r*(a1+a2)/2; grid->D[k] = r*a1; grid->B[k] = (a1+a2)/2*b; grid->H[k] = a1*h; grid->C[k] = -(grid->G[k]+grid->D[k]+a1*h0+(a1+a2)/2*b0); grid->F[k] = dy2*source; 
        }
        else if ((di1==-1)&&(dj2==1)) {
           grid->G[k] = r*a1; grid->D[k] = r*(a1+a2)/2; grid->B[k] = a1*b; grid->H[k] = (a1+a2)/2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+(a1+a2)/2*h0+a1*b0); grid->F[k] = dy2*source; 
        }
        else if ((dj1==-1)&&(di2==-1)) {
            grid->G[k] = r*(a1+a2)/2; grid->D[k] = r*a1; grid->B[k] = a1*b; grid->H[k] = (a1+a2)/2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+(a1+a2)/2*h0+a1*b0); grid->F[k] = dy2*source; 
        }
        else if ((dj1==1)&&(di2==1)) {
             grid->G[k] = r*a1; grid->D[k] = r*(a1+a2)/2; grid->B[k] = (a1+a2)/2*b; grid->H[k] = a1*h; grid->C[k] = -(grid->G[k]+grid->D[k]+a1*h0+(a1+a2)/2*b0); grid->F[k] = dy2*source;  
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;       
}

int _interface_polar(grid_def* grid, int igp, int jgp, int ig, int jg, int ign, int jgn, float source,float a1, float a2) {
     //igp,jgp : previous point
    //ig,jg : point
    //ign,jgn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dtheta2,dr2;
    float h,b,s;
    float dtheta,dr;
    dtheta = grid->dx;
    dr = grid->dy;
    dtheta2=dtheta*dtheta;
    dr2 = dr*dr;
    di1 = ig-igp;
    di2 = ign-ig;
    dj1 = jg-jgp;
    dj2 = jgn-jg;
    if ((ig<0)||(jg<0)||(ig>=grid->width)||(jg>=grid->height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (ig+1)+grid->full_width*(jg+1);
    h = (jg+0.5)*jg*dtheta2;
    b = (jg-0.5)*jg*dtheta2;
    s = jg*jg*dr2*dtheta2;
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            grid->D[k] = grid->G[k] =(a1+a2)/2; grid->H[k] = a1*h; grid->B[k] = a2*b; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source;
        }
        else if (di1==-1) {
            grid->D[k] = grid->G[k] = (a1+a2)/2; grid->H[k] = a2*h; grid->B[k] = a1*b; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source;   
        }
        else if (dj1==1) {
            grid->G[k] = a1; grid->D[k] = a2; grid->H[k] = (a1+a2)/2*h; grid->B[k] = (a1+a2)/2*b; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source; 
        }
        else if (dj1==-1) {
            grid->G[k] = a2; grid->D[k] = a1; grid->H[k] = (a1+a2)/2*h; grid->B[k] = (a1+a2)/2*b; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            grid->G[k] = (a1+a2)/2; grid->D[k] = a2; grid->B[k] = a2*b; grid->H[k] = (a1+a2)/2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source; 
        }
        else if ((di1==-1)&&(dj2==-1)) {
            grid->G[k] = a2; grid->D[k] = (a1+a2)/2; grid->B[k] = (a1+a2)/2*b; grid->H[k] = a2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source;  
        }
        else if ((di2==1)&&(dj1==-1)) {
            grid->G[k] = a2; grid->D[k] = (a1+a2)/2; grid->B[k] = a2*b; grid->H[k] = (a1+a2)/2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
             grid->G[k] = (a1+a2)/2; grid->D[k] = a2; grid->B[k] = (a1+a2)/2*b; grid->H[k] = a2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source; 
        }
        else if ((di1==1)&&(dj2==-1)) {
           grid->G[k] = (a1+a2)/2; grid->D[k] = a1; grid->B[k] = (a1+a2)/2*b; grid->H[k] = a1*h; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source; 
        }
        else if ((di1==-1)&&(dj2==1)) {
           grid->G[k] = a1; grid->D[k] = (a1+a2)/2; grid->B[k] = a1*b; grid->H[k] = (a1+a2)/2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source; 
        }
        else if ((dj1==-1)&&(di2==-1)) {
            grid->G[k] = (a1+a2)/2; grid->D[k] = a1; grid->B[k] = a1*b; grid->H[k] = (a1+a2)/2*h; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source; 
        }
        else if ((dj1==1)&&(di2==1)) {
             grid->G[k] = a1; grid->D[k] = (a1+a2)/2; grid->B[k] = (a1+a2)/2*b; grid->H[k] = a1*h; grid->C[k] = -(grid->G[k]+grid->D[k]+grid->H[k]+grid->B[k]); grid->F[k] = s*source;  
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;    
}

int _interface_polygon(grid_def* grid, int i_start, int j_start, intArray li_list, intArray lj_list, float source, float a1, float a2) {
    int k,nig,njg,ig,jg,ig_p,jg_p,id,jd;
    int ig_start, jg_start, ig_first, jg_first, ign_first, jgn_first;
    ig_p = jg_p = ig_first = jg_first = ign_first = jgn_first = 0;
    if ((lj_list.ndims!=1)||(lj_list.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (li_list.dims[0]!=lj_list.dims[0]) return POISSON_BAD_ARRAY_DIM;
    ig_start = i_start*grid->r;
    jg_start = j_start*grid->r;
    for (k=0; k<li_list.dims[0]; k++) {
        nig = li_list.data[k]*grid->r;
        njg = lj_list.data[k]*grid->r;
        if (nig!=0) {
               jd = 0;
               id = 1;
               if (nig<0) id = -1;
               if (njg!=0) return POISSON_BAD_POLYGON_DEFINITION;
               if (k==0) {
                   ig_first = ig_p = ig_start;
                   jg_first = jg_p = jg_start;
                   ig_start+=id;
                   nig--;
                   ign_first = ig_start;
                   jgn_first = jg_start;
               }
               jg = jg_start;
               for (ig=ig_start; ig!=ig_start+nig; ig+=id) {
                   grid->P[ig+grid->width*jg] = MASK_INTERFACE;
                   solver.interface_function(grid,ig_p, jg_p, ig, jg, ig+id, jg, source, a1, a2);
                   ig_p = ig; jg_p = jg;
                   
               } 
        }
        if (njg!=0) { 
               id = 0;
               jd = 1;
               if (njg<0) jd = -1;
               if (nig!=0) return POISSON_BAD_POLYGON_DEFINITION;
               if (k==0) {
                   ig_first = ig_p = ig_start;
                   jg_first = jg_p = jg_start;
                   jg_start+=jd;
                   njg--;
                   ign_first = ig_start;
                   jgn_first = jg_start;
               }
               ig = ig_start;
               for (jg=jg_start; jg!=jg_start+njg; jg+=jd) {
                   grid->P[ig+grid->width*jg] = MASK_INTERFACE;
                   solver.interface_function(grid,ig_p, jg_p, ig, jg, ig, jg+jd, source, a1, a2);
                   ig_p = ig; jg_p = jg;
                  
               }
        }
        
        ig_start = ig;
        jg_start = jg;
    }
    grid->P[ig_first+grid->width*jg_first] = MASK_INTERFACE;
    solver.interface_function(grid, ig_p, jg_p, ig_first, jg_first, ign_first, jgn_first, source, a1, a2);
    return 0;  
}



int poissonMG_interface_polygon(int i_start, int j_start, intArray li_list, intArray lj_list, float source, float a1, float a2) {
    int g,err;
    for (g=0; g<solver.grid_levels; g++) {
        err = _interface_polygon(grid_list[g],i_start,j_start,li_list,lj_list,source,a1,a2);
        if (err<0) return err;
    }
    return 0;  
}

int poissonMG_init_array() {
    int ig,jg,c;
    grid_def* grid;
    grid = solver.fine_grid;
    for (ig=0; ig<grid->width; ig++) {
        for (jg=0; jg<grid->height; jg++) {
            c = (ig+1)+grid->full_width*(jg+1);
            if (grid->GT[c]==1.0) grid->U[c] = 0.0;
        }
    }
    return 0;   
}

int poissonMG_sinus_mode(int mx, int my) {
    int ig,jg,c;
    float fx,fy;
    grid_def* grid;
    grid = solver.fine_grid;
    fx = (float)(3.1415926*mx/grid->width);
    fy = (float)(3.1415926*my/grid->height);
    for (ig=0; ig<grid->width; ig++) {
        for (jg=0; jg<grid->height; jg++) {
            c = (ig+1)+grid->full_width*(jg+1);
            if (grid->GT[c]==1.0) grid->U[c] += (float)(sin(fx*ig)*sin(fy*jg));
        }
    }
    return 0;
}

int poissonMG_step_mode() {
    int ig,jg;
    int jEnd,c;
    grid_def* grid;
    grid = solver.fine_grid;
    jEnd = grid->height/2;
    for (ig=0; ig<grid->width; ig++) {
        for (jg=0; jg<jEnd; jg++) {
            c = (ig+1)+grid->full_width*(jg+1);
            if (grid->GT[c]==1.0) grid->U[c] = 1.0;
        }
        for (jg=jEnd; jg<grid->height; jg++) {
            c = (ig+1)+grid->full_width*(jg+1);
            if (grid->GT[c]==1.0) grid->U[c] = 0.0;
        }
    }
    return 0;
}

float _array_norm(grid_def* grid) {
    int ig,jg,c,width,height;
    float norm;
    float u;
    float *U;
    norm = 0.0;
    width = grid->width;
    height = grid->height;
    U = grid->U;
    c = 1+grid->full_width;
    for (jg=0; jg<width; jg++) {
        for (ig=0; ig<height; ig++) {
            u = U[c];
            norm += u*u;
            c += 1;
        }
        c += 2;
    }
    return sqrt(norm/grid->size);
}

int poissonMG_get_finegrid_array_norm(float *value) {
    *value =  _array_norm(solver.fine_grid);
    return 0;
}

void _iterationPaire(float *U, float *H, float *B, float *G, float *D, float *C, float *F, int width, int full_width,int jStart, int jEnd, float w1, float w) {
    int ig,jg,c;
    c = 1+full_width*(jStart+1);
    for (jg=jStart; jg<jEnd; jg++) { 
        for (ig=0; ig<width; ig++) {
            //if ((((ig&1)+(jg&1))&1)==0) U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
            if ((c&1)==0) U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
            c += 1;
            
        }
        c += 2;
    }
}

void _iterationImpaire(float *U, float *H, float *B, float *G, float *D, float *C, float *F, int width, int full_width,int jStart, int jEnd, float w1, float w) {
    int ig,jg,c;
    c = 1+full_width*(jStart+1);
    for (jg=jStart; jg<jEnd; jg++) {
        for (ig=0; ig<width; ig++) {
            //if ((((ig&1)+(jg&1))&1)==1) U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
            if ((c&1)==1) U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
            c += 1;
        }
        c += 2;
    }
}

void _iterations_redblack(grid_def* grid, int niter, float w) {
    int i;
    int width, full_width,height;
    float *U,*H,*B,*G,*D,*C,*F;
    float w1;
    w1 = 1.0-w;
    width = grid->width;
    height = grid->height;
    full_width = grid->full_width;
    U = grid->U; H = grid->H; B = grid->B; G = grid->G; D = grid->D; C = grid->C; F = grid->F;
    for (i=0; i<niter; i++) {
        _iterationPaire(U,H,B,G,D,C,F,width,full_width,0,height,w1,w);
        _iterationImpaire(U,H,B,G,D,C,F,width,full_width,0,height,w1,w);
        
    }
}

void _iterations(grid_def* grid, int niter, float w) {
    int i,ig,jg,c;
    int width, full_width,height;
    float *U,*H,*B,*G,*D,*C,*F;
    float w1;
    w1 = 1.0-w;
    width = grid->width;
    height = grid->height;
    full_width = grid->full_width;
    U = grid->U; H = grid->H; B = grid->B; G = grid->G; D = grid->D; C = grid->C; F = grid->F;
    
    for (i=0; i<niter; i++) {
        c = 1+full_width;
        for (jg=0; jg<height; jg++) {
                for (ig=0; ig<width; ig++) {
                    U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
                    c += 1;
                }
                c += 2;
        }
        
    }
}

void _iteration_polar(float *U, float *H, float *B, float *G, float *D, float *C, float *F, int width, int full_width, int jg_min, int height, float dtheta, float w1, float w) {
    int ig,jg,c,c0,j0;
    float sum;
    c0 = 1+full_width; // centre (jg=0)
    j0 = jg_min;
    if (j0 < 2) j0 = 2;
    if (jg_min<1) {
    if (C[c0]!=1.0) {
        sum = 0;
        c = 1+full_width*2;
        for (ig=0; ig<width; ig++) {
            sum += U[c];
            c += 1;
        }
        U[c0] = (sum-F[c0]*2/dtheta)/width;
    }
    }
    if (jg_min<2) {
    c = 1+full_width*2; //premier cercle (jg=1)
    U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c+width-1]-H[c]*U[c+full_width]-B[c]*U[c0])/C[c];
    c += 1;
    for (ig=1; ig<width-1; ig++) {
         U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c0])/C[c];
         c += 1;
    }
    U[c] = w1*U[c]+w*(F[c]-D[c]*U[c-width+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c0])/C[c];
    c += 3;
    }
    c = 1+full_width*(j0+1);
    for (jg=j0; jg<height; jg++) {
        // premier point de la ligne
        U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c+width-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];  
        c += 1;
        for (ig=1; ig<width-1; ig++) {
            U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
            c += 1;
        }
       //dernier point de la ligne
        U[c] = w1*U[c]+w*(F[c]-D[c]*U[c-width+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
        c += 3;
    }
    // traitement du centre
    
}

void _iterationPaire_polar(float *U, float *H, float *B, float *G, float *D, float *C, float *F, int width, int full_width,int jg_min, int height, float dtheta, float w1, float w) {
    int ig,jg,c,c0,j0;
    float sum;
    j0 = jg_min;
    if (j0 < 2) j0 = 2;
    c0 = 1+full_width; // centre
    if (jg_min<1) {
    // traitement du centre (jg=0)
    if (C[c0]!=1.0) {
        sum = 0;
        c = 1+full_width*2;
        for (ig=0; ig<width; ig++) { // somme du premier cercle
            sum += U[c];
            c += 1;
        }
        U[c0] = (sum-F[c0]*2/dtheta)/width;
    }
    }
    if (jg_min<2) {
    c = 1+full_width*2; // premier cercle (jg=1)
    if ((c&1)==0) U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c+width-1]-H[c]*U[c+full_width]-B[c]*U[c0])/C[c];
    c += 1;
    for (ig=1; ig<width-1; ig++) {
         if ((c&1)==0) U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c0])/C[c];
         c += 1;
    }
    if ((c&1)==0) U[c] = (F[c]-D[c]*U[c-width+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c0])/C[c];
    c += 3;
    }
    c = 1+full_width*(j0+1);
    for (jg=j0; jg<height; jg++) {
        if ((c&1)==0) { // premier point de la ligne
             U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c+width-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];  
        }
        c += 1;
        for (ig=1; ig<width-1; ig++) {
            if ((c&1)==0) U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
            c += 1;
        }
        if ((c&1)==0) { //dernier point de la ligne
            U[c] = (F[c]-D[c]*U[c-width+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
        }
        c += 3;
    }
    
}

void _iterationImpaire_polar(float *U, float *H, float *B, float *G, float *D, float *C, float *F, int width, int full_width,int jg_min, int height, float dtheta, float w1, float w) {
    int ig,jg,c,c0,j0;
    c0 = 1+full_width; // centre
    j0 = jg_min;
    if (j0 < 2) j0 = 2;
    if (jg_min<2) {
    c = 1+full_width*2; //premier cercle (jg=1)
    if ((c&1)==1) U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c+width-1]-H[c]*U[c+full_width]-B[c]*U[c0])/C[c];
    c += 1;
    for (ig=0; ig<width; ig++) {
         if ((c&1)==1) U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c0])/C[c];
         c += 1;
    }
    if ((c&1)==1) U[c] = (F[c]-D[c]*U[c-width+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c0])/C[c];
    c += 3;
    }
    c = 1+full_width*(j0+1);
    for (jg=j0; jg<height; jg++) {
        if ((c&1)==1) { // premier point de la ligne
             U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c+width-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];  
        }
        c += 1;
        for (ig=1; ig<width-1; ig++) {
            if ((c&1)==1) U[c] = (F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
            c += 1;
        }
        if ((c&1)==1) { //dernier point de la ligne
            U[c] = (F[c]-D[c]*U[c-width+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width])/C[c];
        }
        c += 3;
    }
}

void _iterations_polar(grid_def* grid, int niter, float w) {
    int i,jg_min;
    int width, full_width,height;
    float *U,*H,*B,*G,*D,*C,*F;
    float dtheta;
    float w1;
    w1 = 1.0-w;
    width = grid->width;
    height = grid->height;
    full_width = grid->full_width;
    dtheta = grid->dx;
    jg_min = grid->jg_min;
    U = grid->U; H = grid->H; B = grid->B; G = grid->G; D = grid->D; C = grid->C; F = grid->F;
    for (i=0; i<niter; i++) {
        _iteration_polar(U,H,B,G,D,C,F,width,full_width,jg_min,height,dtheta,w1,w);
    }
}

void* _iterations_thread(void *arg) {
    int width, full_width;
    float *U,*H,*B,*G,*D,*C,*F;
    threadArg* ta = (threadArg*)arg;
    int niter = ta->niter;
    int jStart = ta->jStart;
    int jEnd = ta->jEnd;
    float w1 = ta->w1;
    float w = ta->w;
    int i,r;
    grid_def *grid = ta->grid;
    width = grid->width;
    full_width = grid->full_width;
    U = grid->U; H = grid->H; B = grid->B; G = grid->G; D = grid->D; C = grid->C; F = grid->F;
    for (i=0; i<niter; i++) {
         _iterationPaire(U,H,B,G,D,C,F,width,full_width,jStart,jEnd,w1,w);
         r =  pthread_barrier_wait(&threadBarrier);
         _iterationImpaire(U,H,B,G,D,C,F,width,full_width,jStart,jEnd,w1,w);
         r =  pthread_barrier_wait(&threadBarrier);
         if(r == PTHREAD_BARRIER_SERIAL_THREAD) {
             
         }
         
    }
    pthread_exit(NULL);
}


int _iterations_2threads(grid_def* grid, int niter, float w) { // CPU iteration, 2 threads
    pthread_attr_t attr;
    pthread_t thread1, thread2;
    threadArg arg1, arg2;
    int err;
    float w1;
    int size;
    w1 = 1.0-w;
    
    size = (grid->height-1)/2;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    err = pthread_barrier_init(&threadBarrier,NULL,2);
    if (err) return POISSON_ERROR_THREAD_BARRIER;
    arg1.grid = grid;
    arg1.niter = niter;
    arg1.jStart = 0;
    arg1.jEnd = size;
    arg1.w1 = w1;
    arg1.w = w;
    arg2.grid = grid;
    arg2.niter = niter;
    arg2.jStart = size;
    arg2.jEnd = grid->height;
    arg2.w1 = w1;
    arg2.w = w;
        err = pthread_create(&thread1,&attr,_iterations_thread,(void*)&arg1);
        if (err) return POISSON_ERROR_THREAD_START;
        err = pthread_create(&thread2,&attr,_iterations_thread,(void*)&arg2);
        if (err) return POISSON_ERROR_THREAD_START;
        err = pthread_join(thread1,NULL);
        if (err) return POISSON_ERROR_THREAD_JOIN;
        err = pthread_join(thread2,NULL);
        if (err) return POISSON_ERROR_THREAD_JOIN;
        
    pthread_attr_destroy(&attr);
    pthread_barrier_destroy(&threadBarrier);
    return 0; 
}

int _iterations_4threads(grid_def* grid, int niter, float w) { // CPU iteration, 4 threads
    pthread_attr_t attr;
    pthread_t thread1, thread2, thread3, thread4;
    threadArg arg1, arg2, arg3, arg4;
    int err;
    int size;
    float w1;
    w1 = 1.0-w;
    
    size = (grid->height-1)/4;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    err = pthread_barrier_init(&threadBarrier,NULL,4);
    if (err) return POISSON_ERROR_THREAD_BARRIER;
    arg1.grid = grid;
    arg1.niter = niter;
    arg1.jStart = 0;
    arg1.jEnd = size;
    arg1.w1 = w1;
    arg1.w = w;
    arg2.grid = grid;
    arg2.niter = niter;
    arg2.jStart = size;
    arg2.jEnd = 2*size;
    arg2.w1 = w1;
    arg2.w = w;
    arg3.grid = grid;
    arg3.niter = niter;
    arg3.jStart = 2*size;
    arg3.jEnd = 3*size;
    arg3.w1 = w1;
    arg3.w = w;
    arg4.grid = grid;
    arg4.niter = niter;
    arg4.jStart = 3*size;
    arg4.jEnd = grid->height;
    arg4.w1 = w1;
    arg4.w = w;
        err = pthread_create(&thread1,&attr,_iterations_thread,(void*)&arg1);
        if (err) return POISSON_ERROR_THREAD_START;
        err = pthread_create(&thread2,&attr,_iterations_thread,(void*)&arg2);
        if (err) return POISSON_ERROR_THREAD_START;
        err = pthread_create(&thread3,&attr,_iterations_thread,(void*)&arg3);
        if (err) return POISSON_ERROR_THREAD_START;
        err = pthread_create(&thread4,&attr,_iterations_thread,(void*)&arg4);
        if (err) return POISSON_ERROR_THREAD_START;
        err = pthread_join(thread1,NULL);
        if (err) return POISSON_ERROR_THREAD_JOIN;
        err = pthread_join(thread2,NULL);
        if (err) return POISSON_ERROR_THREAD_JOIN;
        err = pthread_join(thread3,NULL);
        if (err) return POISSON_ERROR_THREAD_JOIN;
        err = pthread_join(thread4,NULL);
        if (err) return POISSON_ERROR_THREAD_JOIN;
    pthread_attr_destroy(&attr);
    pthread_barrier_destroy(&threadBarrier);
    return 0; 
}

int poissonMG_finegrid_iterations(int niter, float w) {
    solver.iterations_function(solver.fine_grid,niter,w);
    return 0;
}

int poissonMG_finegrid_iterations_2threads(int niter, float w) {
    _iterations_2threads(solver.fine_grid,niter,w);
    return 0;
}

int poissonMG_finegrid_iterations_4threads(int niter, float w) {
    _iterations_4threads(solver.fine_grid,niter,w);
    return 0;
}



void _residu(grid_def* grid) {
    int ig,jg,c,width, full_width,height;
    float *U,*H,*B,*G,*D,*C,*F,*R;
    width = grid->width;
    height = grid->height;
    full_width = grid->full_width;
    U = grid->U; H = grid->H; B = grid->B; G = grid->G; D = grid->D; C = grid->C; F = grid->F; R = grid->R;
    c = 1+full_width;
    for (jg=0; jg<height; jg++) {
        for (ig=0; ig<width; ig++) {
            R[c] = F[c]-(D[c]*U[c+1]+G[c]*U[c-1]+H[c]*U[c+full_width]+B[c]*U[c-full_width]+C[c]*U[c]);
            //if (jg==1) printf("%g,%g,%g,%g,%g,%g,%g\n",C[c],D[c],G[c],H[c],B[c],F[c],R[c]);
            c+=1;
        }
        c += 2;
    }
    
}

void _residu_polar(grid_def* grid) {
    int ig,jg,jg_min,j0,c,c0,width, full_width,height;
    float *U,*H,*B,*G,*D,*C,*F,*R,sum,dtheta;
    dtheta = grid->dx;
    width = grid->width;
    height = grid->height;
    full_width = grid->full_width;
    U = grid->U; H = grid->H; B = grid->B; G = grid->G; D = grid->D; C = grid->C; F = grid->F; R = grid->R;
    jg_min = grid->jg_min;
    if (jg_min<1) {
    c0 =1+full_width;
    c = 1+full_width*2;
    sum = 0.0;
    for (ig=1; ig<=width; ig++) {
            sum += U[c];
            c += 1;
        }
    R[c0] = (F[c0]*2/dtheta-sum)+U[c0]*width;
    }
    j0 = jg_min;
    if (j0==0) j0=1;
    c = 1+full_width*(j0+1);
    for (jg=j0; jg<height; jg++) {
        R[c] = F[c]-D[c]*U[c+1]-G[c]*U[c+width-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width]-C[c]*U[c];
        c+=1;
        for (ig=1; ig<width-1; ig++) {
            R[c] = F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width]-C[c]*U[c];
            c+=1;
        }
        R[c] = F[c]-D[c]*U[c-width+1]-G[c]*U[c-1]-H[c]*U[c+full_width]-B[c]*U[c-full_width]-C[c]*U[c];
        c+=3;
    }
    
}

int poissonMG_get_array(floatArray *array, int symetry) {
    int ig,jg,k,k0;
    int offset;
    grid_def *grid;
    grid = solver.fine_grid;
    if (array->ndims!=2) return -255;
    if (symetry) offset = grid->width*(grid->height-1); 
    else offset = 0;    
    for (jg=0; jg< grid->height; jg++) {
        for (ig=0; ig< grid->width; ig++) {
            k = (ig+1)+grid->full_width*(jg+1);
            grid->U1[offset+ig+jg*grid->width] = grid->U[k];
            if ((symetry)&&(jg!=0)) grid->U1[offset+ig-jg*grid->width] = symetry*grid->U[k];
            
        }
    }
    if (solver.type==TYPE_POLAR) {
        jg=0;
        k0 = 1+grid->full_width;
        for (ig=1; ig< grid->width; ig++) {
               grid->U1[offset+ig+jg*grid->width] = grid->U[k0];
        }
    }
    array->dims[1] = grid->width;
    array->dims[0] = grid->height;
    if (symetry) array->dims[0] = 2*grid->height-1;
    array->data = grid->U1;
    return 0;   
}

void _fill_mask(grid_def *grid) {
    int ig,jg,k;
    char fill;
    for (jg=0; jg< grid->height; jg++) {
        fill = 0;
        for (ig=1; ig< grid->width-1; ig++) {
                 k = ig+grid->width*jg;
                 if ((fill==1)&&(grid->P[k]==MASK_INSIDE)) {
                     grid->P[k] = MASK_OUTSIDE;   
                 }
                 if ((grid->P[k]==MASK_NEUMANN)&&(grid->P[k-1]!=MASK_NEUMANN)&&(grid->P[k+1]!=MASK_NEUMANN)) {
                    fill = !fill;   
                 }
        }
    }
}

int poissonMG_get_mask(byteArray *array, int symetry) {
    int ig,jg,k,k0;
    int offset;
    int v;
    grid_def *grid;
    grid = solver.fine_grid;
    if (array->ndims!=2) return -255;
    _fill_mask(grid);
    if (symetry) offset = grid->width*(grid->height-1); 
    else offset = 0;    
    for (jg=0; jg< grid->height; jg++) {
        for (ig=0; ig< grid->width; ig++) {
            k = ig+grid->width*(grid->height-1-jg);
            if (grid->P[k]!=MASK_INSIDE) v = 255;
            else v  =0;
            grid->M[ig+jg*grid->width] = v;
            if ((symetry)&&(jg!=0)) grid->M[offset+ig+(grid->height-1-jg)*grid->width] = v;
        }
    }
    if (solver.type==TYPE_POLAR) {
        jg=0;
        k0 = 0+grid->width*(grid->height-1-jg);
        for (ig=1; ig< grid->width; ig++) {
               if (grid->P[k0]!=MASK_INSIDE) grid->M[ig+jg*grid->width] = 255;
        }
    }
    array->dims[1] = grid->width;
    array->dims[0] = grid->height;
    if (symetry) array->dims[0] = 2*grid->height-1;
    array->data = grid->M;
    
    return 0;   
}

int poissonMG_get_derivX(floatArray *array, int symetry) {
    int ig,jg,k;
    int offset;
    float dx;
    grid_def *grid;
    grid = solver.fine_grid;
    dx = grid->dx;
    if (array->ndims!=2) return -255;
    if (symetry) offset = grid->width*(grid->height-1);  
    else offset = 0;
    ig = 0;
    for (jg=0; jg<grid->height; jg++) {
        k = (ig+1)+grid->full_width*(jg+1);
        grid->U1[offset+ig+grid->width*jg] = (grid->U[k+1]-grid->U[k])/dx;
        
    }
    for (ig=1; ig<grid->width-1; ig++) {
        for (jg=0; jg<grid->height; jg++) {
            k = (ig+1)+grid->full_width*(jg+1);
            grid->U1[offset+ig+grid->width*jg] = (grid->U[k+1]-grid->U[k-1])/(2*dx);
            if (symetry) grid->U1[offset+ig-grid->width*jg] = symetry*grid->U1[offset+ig+grid->width*jg];
        }   
    }
    ig = grid->width-1;
    for (jg=0; jg<grid->height; jg++) {
        k = (ig+1)+grid->full_width*(jg+1);
        grid->U1[offset+ig+grid->width*jg] = (grid->U[k]-grid->U[k-1])/dx;
        if (symetry) grid->U1[offset+ig-grid->width*jg] = symetry*grid->U1[offset+ig+grid->width*jg];
    }
    array->dims[1] = grid->width;
    array->dims[0] = grid->height;
    if (symetry) array->dims[0] = 2*grid->height-1;
    array->data = grid->U1;
    return 0;    
}

int poissonMG_get_derivY(floatArray *array,int symetry) {
    int ig,jg,k;
    int offset;
    float dy;
    grid_def *grid;
    grid = solver.fine_grid;
    dy = grid->dy;
    if (array->ndims!=2) return -255;
    if (symetry) offset = grid->width*(grid->height-1);  
    else offset = 0;
    jg=0;
    for (ig=0; ig<grid->width; ig++) {
        k = (ig+1)+grid->full_width*(jg+1);
        grid->U1[offset+ig+grid->width*jg] = (grid->U[k+grid->full_width]-grid->U[k])/dy;
        
    }
    for (jg=1; jg<grid->height-1; jg++) {
        for (ig=0; ig<grid->width; ig++) {
            k = (ig+1)+grid->full_width*(jg+1);
            grid->U1[offset+ig+grid->width*jg] = (grid->U[k+grid->full_width]-grid->U[k-grid->full_width])/(2*dy);
            if (symetry) grid->U1[offset+ig-grid->width*jg] = symetry*grid->U1[offset+ig+grid->width*jg];
        }
    }
    jg= grid->height-1;
    for (ig=0; ig<grid->width; ig++) {
        k = (ig+1)+grid->full_width*(jg+1);
        grid->U1[offset+ig+grid->width*jg] = (grid->U[k]-grid->U[k-grid->full_width])/dy;
        if (symetry) grid->U1[offset+ig-grid->width*jg] = symetry*grid->U1[offset+ig+grid->width*jg];
    }
    array->dims[1] = grid->width;
    array->dims[0] = grid->height;
    if (symetry) array->dims[0] = 2*grid->height-1;
    array->data = grid->U1;
    return 0;   
}

int poissonMG_get_derivRUR_axial(floatArray *array, int symetry) {
    int ig,jg,k;
    int offset;
    float dr;
    grid_def *grid;
    grid = solver.fine_grid;
    dr = grid->dy;
    if (array->ndims!=2) return -255;
    if (symetry) offset = grid->width*(grid->height-1);  
    else offset = 0;
    jg = 0;
    for (ig=0; ig<grid->width; ig++) {
        k=(ig+1)+grid->full_width*(jg+1);
        grid->U1[offset+ig+grid->width*jg] = ((jg+2)*grid->U[k+grid->full_width]-(jg+1)*grid->U[k])/((jg+1)*dr);
        
    }
    for (jg=1; jg<grid->height-1; jg++) {
        for (ig=0; ig<grid->width; ig++) {
            k=(ig+1)+grid->full_width*(jg+1);
            grid->U1[offset+ig+grid->width*jg] = ((jg+1)*grid->U[k+grid->full_width]-(jg-1)*grid->U[k-grid->full_width])/(2*jg*dr);
            if (symetry) grid->U1[offset+ig-grid->width*jg] = symetry*grid->U1[offset+ig+grid->width*jg];
        }   
    }
    jg=grid->height-1;
    for (ig=0; ig<grid->width; ig++) {
        k = (ig+1)+grid->full_width*(jg+1);
        grid->U1[offset+ig+grid->width*jg] = (jg*grid->U[k]-(jg-1)*grid->U[k-grid->full_width])/(jg*dr);
        if (symetry) grid->U1[offset+ig-grid->width*jg] = symetry*grid->U1[offset+ig+grid->width*jg];
    }
    array->dims[1] = grid->width;
    array->dims[0] = grid->height;
    if (symetry) array->dims[0] = 2*grid->height-1;
    array->data = grid->U1;
    return 0; 
}

int poissonMG_get_RUR_axial(floatArray *array, int symetry) {
    int ig,jg,k;
    int offset;
    float dr;
    grid_def *grid;
    grid = solver.fine_grid;
    dr = grid->dy;
    if (array->ndims!=2) return -255;
    if (symetry) offset = grid->width*(grid->height-1);  
    else offset = 0;
    for (jg=0; jg<grid->height-1; jg++) {
        for (ig=0; ig<grid->width; ig++) {
             k=(ig+1)+grid->full_width*(jg+1);
             grid->U1[offset+ig+grid->width*jg] = jg*dr*grid->U[k];
             if (symetry) grid->U1[offset+ig-grid->width*jg] = symetry*jg*dr*grid->U[k];
        }
    }
    array->dims[1] = grid->width;
    array->dims[0] = grid->height;
    if (symetry) array->dims[0] = 2*grid->height-1;
    array->data = grid->U1;
    return 0;
}

int poissonMG_get_derivTheta_polar(floatArray *array) {
    int ig,jg,k;
    float dtheta,r,dr;
    grid_def *grid;
    grid = solver.fine_grid;
    dtheta = grid->dx;
    dr = grid->dy;
    if (array->ndims!=2) return -255;
    jg = 0;
    for (ig=0; ig<grid->width; ig++) {
        k = (ig+1)+grid->full_width*(jg+1);
        grid->U1[ig+grid->width*jg] = 0.0;
    }
    for (jg=1; jg<grid->height; jg++) {
        ig = 0;
        k = (ig+1)+grid->full_width*(jg+1);
        r = dr*jg;
        grid->U1[ig+grid->width*jg] = (grid->U[k+1]-grid->U[k+grid->width-1])/(2*dtheta)/r;
        for (ig=1; ig<grid->width-1; ig++) {
             k = (ig+1)+grid->full_width*(jg+1);
             grid->U1[ig+grid->width*jg] = (grid->U[k+1]-grid->U[k-1])/(2*dtheta)/r;  
        }
        k = (ig+1)+grid->full_width*(jg+1);
        grid->U1[ig+grid->width*jg] = (grid->U[k-grid->width+1]-grid->U[k-1])/(2*dtheta)/r; 
    }
    array->dims[1] = grid->width;
    array->dims[0] = grid->height;
    array->data = grid->U1;
    return 0; 
       
}

float _residu_norm(grid_def* grid) {
    int ig,jg,c,width,height;
    float norm;
    float r;
    float *R;
    norm = 0.0;
    width = grid->width;
    height = grid->height;
    R = grid->R;
    c = 1+grid->full_width;
    for (jg=0; jg<width; jg++) {
        for (ig=0; ig<height; ig++) {
            r = R[c];
            norm += r*r;
            c += 1;
        }
        c += 2;
    }
    return sqrt(norm/grid->size);
}

int poissonMG_get_finegrid_residu_norm(float *value) {
    _residu(solver.fine_grid);
    *value =  _residu_norm(solver.fine_grid);
    return 0;
}



void __linear_prolongation(grid_def *grid_2h, grid_def *grid_h) {
    int ig_2h, jg_2h, width_2h, height_2h, c_2h, c_h;
    float *U_h, *U_2h;
    char *GT_h;
    width_2h = grid_2h->width;
    height_2h = grid_2h->height;
    c_2h = 1+grid_2h->full_width;
    c_h = 1+grid_h->full_width;
    U_h = grid_h->U;
    U_2h = grid_2h->U;
    GT_h = grid_h->GT;
    for (jg_2h=0; jg_2h<height_2h-1; jg_2h++) {
        for (ig_2h=0; ig_2h<width_2h-1; ig_2h++) {
            U_h[c_h] += GT_h[c_h]*U_2h[c_2h];
            U_h[c_h+1] += GT_h[c_h+1]*0.5*(U_2h[c_2h]+U_2h[c_2h+1]);
            U_h[c_h+grid_h->full_width] += GT_h[c_h+grid_h->full_width]*0.5*(U_2h[c_2h]+U_2h[c_2h+grid_2h->full_width]);
            U_h[c_h+grid_h->full_width+1] += GT_h[c_h+grid_h->full_width+1]*0.25*(U_2h[c_2h]+U_2h[c_2h+1]+U_2h[c_2h+grid_2h->full_width]+U_2h[c_2h+grid_2h->full_width+1]);
            c_h += 2;
            c_2h += 1;
        }
        c_2h += 3;
        c_h += 3+grid_h->full_width;
    }
    c_h = 2*grid_h->full_width-2;
    c_2h = 2*grid_2h->full_width-2;
    for (jg_2h=0; jg_2h<height_2h-1; jg_2h++) { // dernière colonne
            U_h[c_h] += GT_h[c_h]*U_2h[c_2h];
            U_h[c_h+grid_h->full_width] += GT_h[c_h+grid_h->full_width]*0.5*(U_2h[c_2h]+U_2h[c_2h+grid_2h->full_width]);
            c_2h += grid_2h->full_width;
            c_h += 2*grid_h->full_width;
    }
    c_h = 1+grid_h->full_width*grid_h->height;
    c_2h = 1+grid_2h->full_width*height_2h;
    for (ig_2h=0; ig_2h<width_2h-1; ig_2h++) { // dernière ligne
        U_h[c_h] += GT_h[c_h]*U_2h[c_2h];
        U_h[c_h+1] += GT_h[c_h+1]*0.5*(U_2h[c_2h]+U_2h[c_2h+1]);
        c_2h +=1;
        c_h += 2;
    }
    U_h[c_h] += GT_h[c_h]*U_2h[c_2h]; // coin supérieur droit
}

void _linear_prolongation(grid_def *grid_2h, grid_def *grid_h) {
    int ig_2h, jg_2h, ig_h, jg_h, width_2h, height_2h, width_h, height_h, c_h;
    float *U_h, *U_2h, *R_h;
    char *GT_h;
    width_2h = grid_2h->width;
    height_2h = grid_2h->height;
    width_h = grid_h->width;
    height_h = grid_h->height;
    U_h = grid_h->U;
    U_2h = grid_2h->U;
    GT_h = grid_h->GT;
    R_h = grid_h->R;
    /*for (jg_h=0; jg_h<height_h; jg_h++) {
            for (ig_h=0; ig_h<width_h; ig_h++) {
                   c_h = ig_h+1+grid_h->full_width*(jg_h+1);
                   R_h[c_h] = 0.0;
            }
    }*/
    for (jg_2h=0; jg_2h<height_2h; jg_2h++) {
        for (ig_2h=0; ig_2h<width_2h; ig_2h++) {
            R_h[2*ig_2h+1+grid_h->full_width*(2*jg_2h+1)] = U_2h[ig_2h+1+grid_2h->full_width*(jg_2h+1)];
            
        }
    }
    for (ig_h=1; ig_h<width_h; ig_h+=2){
            for (jg_h=0; jg_h<height_h; jg_h+=2) {
                c_h = ig_h+1+grid_h->full_width*(jg_h+1);
                
                R_h[c_h] = 0.5*(R_h[c_h-1]+R_h[c_h+1]);
                
            }
        }
    for (jg_h=1; jg_h<height_h; jg_h+=2) {
            for (ig_h=0; ig_h<width_h; ig_h++) {
                 c_h = ig_h+1+grid_h->full_width*(jg_h+1);
                 R_h[c_h] = 0.5*(R_h[c_h+grid_h->full_width]+R_h[c_h-grid_h->full_width]);
                 
            }
        }
    for (jg_h=0; jg_h<height_h; jg_h++) {
            for (ig_h=0; ig_h<width_h; ig_h++) {
                   c_h = ig_h+1+grid_h->full_width*(jg_h+1);
                   U_h[c_h] += GT_h[c_h]*R_h[c_h];
                   
            }
        }
        
    
}

void _linear_prolongation_polar(grid_def *grid_2h, grid_def *grid_h) {
    int ig_2h, jg_2h, jg_2h_min,j0,width_2h, height_2h, c_2h, c_h, c0_h, c0_2h;
    float *U_h, *U_2h;
    char *GT_h;
    width_2h = grid_2h->width;
    height_2h = grid_2h->height;
    U_h = grid_h->U;
    U_2h = grid_2h->U;
    GT_h = grid_h->GT;
    jg_2h_min = grid_2h->jg_min;
    if (jg_2h_min<1) {
        c_2h = c0_2h = 1+grid_2h->full_width;
        c_h = c0_h = 1+grid_h->full_width;
        jg_2h = 0; // première ligne = centre
        for (ig_2h=0; ig_2h<width_2h-1; ig_2h++) {
            U_h[c_h] += GT_h[c_h]*U_2h[c_2h];
            U_h[c_h+grid_h->full_width] += GT_h[c_h+grid_h->full_width]*0.5*(U_2h[c0_2h]+U_2h[c_2h+grid_2h->full_width]);
            U_h[c_h+grid_h->full_width+1] += GT_h[c_h+grid_h->full_width+1]*1.0/3*(U_2h[c0_2h]+U_2h[c_2h+grid_2h->full_width]+U_2h[c_2h+grid_2h->full_width+1]);
            c_2h +=1;
            c_h +=2;
         
        }
        U_h[c_h] += GT_h[c_h]*U_2h[c_2h];
        U_h[c_h+grid_h->full_width] += GT_h[c_h+grid_h->full_width]*0.5*(U_2h[c0_2h]+U_2h[c_2h+grid_2h->full_width-width_2h+1]);
        U_h[c_h+grid_h->full_width+1] += GT_h[c_h+grid_h->full_width+1]*1.0/3*(U_2h[c0_2h]+U_2h[c_2h+grid_2h->full_width]+U_2h[c_2h+grid_2h->full_width-width_2h+1]);
        /*c_2h +=1;
        c_h +=2;
        c_2h += 2;
        c_h += 2+grid_h->full_width;*/
    }
    j0 = jg_2h_min;
    if (j0==0) j0 = 1;
    c_2h = c0_2h = 1+grid_2h->full_width*(j0+1);
    c_h = c0_h = 1+grid_h->full_width*(2*j0+1);
    for (jg_2h=j0; jg_2h<height_2h-1; jg_2h++) {
        for (ig_2h=0; ig_2h<width_2h-1; ig_2h++) {
            U_h[c_h] += GT_h[c_h]*U_2h[c_2h];
            U_h[c_h+1] += GT_h[c_h+1]*0.5*(U_2h[c_2h]+U_2h[c_2h+1]);
            U_h[c_h+grid_h->full_width] += GT_h[c_h+grid_h->full_width]*0.5*(U_2h[c_2h]+U_2h[c_2h+grid_2h->full_width]);
            U_h[c_h+grid_h->full_width+1] += GT_h[c_h+grid_h->full_width+1]*0.25*(U_2h[c_2h]+U_2h[c_2h+1]+U_2h[c_2h+grid_2h->full_width]+U_2h[c_2h+grid_2h->full_width+1]);
            c_2h += 1;
            c_h += 2;
            
        }
        U_h[c_h] += GT_h[c_h]*U_2h[c_2h];
        U_h[c_h+1] += GT_h[c_h+1]*0.5*(U_2h[c_2h]+U_2h[c_2h-width_2h+1]);
        U_h[c_h+grid_h->full_width] += GT_h[c_h+grid_h->full_width]*0.5*(U_2h[c_2h]+U_2h[c_2h+grid_2h->full_width]);
        U_h[c_h+grid_h->full_width+1] += GT_h[c_h+grid_h->full_width+1]*0.25*(U_2h[c_2h]+U_2h[c_2h-width_2h+1]+U_2h[c_2h+grid_2h->full_width]+U_2h[c_2h+grid_2h->full_width-width_2h+1]);
        c_2h += 1;
        c_h += 2;
        c_2h += 2;
        c_h += 2+grid_h->full_width;
    }
    c_h = 1+grid_h->full_width*grid_h->height;
    c_2h = 1+grid_2h->full_width*height_2h;
    for (ig_2h=0; ig_2h<width_2h-1; ig_2h++) { // dernière ligne
        U_h[c_h] += GT_h[c_h]*U_2h[c_2h];
        U_h[c_h+1] += GT_h[c_h+1]*0.5*(U_2h[c_2h]+U_2h[c_2h+1]);
        c_2h +=1;
        c_h += 2;
    }
    U_h[c_h] += GT_h[c_h]*U_2h[c_2h];
    U_h[c_h+1] += GT_h[c_h+1]*0.5*(U_2h[c_2h]+U_2h[c_2h-width_2h+1]);
   
} 

void _fullweighting_restriction(grid_def *grid_h, grid_def *grid_2h) {
    int ig_2h, jg_2h, width_2h, height_2h, fullwidth_h, c_2h, c_h;
    float *F_2h, *U_2h,  *R_h;
    char *GT_2h;
    width_2h = grid_2h->width;
    height_2h = grid_2h->height;
    fullwidth_h = grid_h->full_width;
    c_2h = 1+grid_2h->full_width;
    c_h = 1+grid_h->full_width;
    F_2h = grid_2h->F;
    R_h = grid_h->R;
    U_2h = grid_2h->U;
    GT_2h = grid_2h->GT;
    for (jg_2h=0; jg_2h<height_2h; jg_2h++) {
        for (ig_2h=0; ig_2h<width_2h; ig_2h++) {
            if ((ig_2h==0)||(ig_2h==width_2h-1)||(jg_2h==0)||(jg_2h==height_2h-1))
                F_2h[c_2h] = 4*GT_2h[c_2h]*R_h[c_h];
            else
            F_2h[c_2h] = 4*GT_2h[c_2h]*(0.0625*(R_h[c_h-fullwidth_h-1]+R_h[c_h-fullwidth_h+1]+R_h[c_h+fullwidth_h-1]+R_h[c_h+fullwidth_h+1])+0.125*(R_h[c_h-fullwidth_h]+R_h[c_h+fullwidth_h]+R_h[c_h+1]+R_h[c_h-1])+0.25*R_h[c_h]);
            U_2h[c_2h] = 0.0;
            c_2h +=1;
            c_h += 2;
        }
        c_2h += 2;
        c_h += 1+grid_h->full_width;
    }
}

void _halfweighting_restriction(grid_def *grid_h, grid_def *grid_2h) {
    int ig_2h, jg_2h, width_2h, height_2h, fullwidth_h, c_2h, c_h;
    float *F_2h, *U_2h,  *R_h;
    char *GT_2h;
    width_2h = grid_2h->width;
    height_2h = grid_2h->height;
    fullwidth_h = grid_h->full_width;
    c_2h = 1+grid_2h->full_width;
    c_h = 1+grid_h->full_width;
    F_2h = grid_2h->F;
    R_h = grid_h->R;
    U_2h = grid_2h->U;
    GT_2h = grid_2h->GT;
    for (jg_2h=0; jg_2h<height_2h; jg_2h++) {
        for (ig_2h=0; ig_2h<width_2h; ig_2h++) {
            if ((ig_2h==0)||(ig_2h==width_2h-1)||(jg_2h==0)||(jg_2h==height_2h-1))
                F_2h[c_2h] = 4*GT_2h[c_2h]*R_h[c_h];
            else
            F_2h[c_2h] = 4*GT_2h[c_2h]*(0.125*(R_h[c_h-fullwidth_h]+R_h[c_h+fullwidth_h]+R_h[c_h+1]+R_h[c_h-1])+0.5*R_h[c_h]);
            U_2h[c_2h] = 0.0;
            c_2h +=1;
            c_h += 2;
        }
        c_2h += 2;
        c_h += 1+grid_h->full_width;
    }
}



void _injection_restriction(grid_def *grid_h, grid_def *grid_2h) {
    int ig_2h, jg_2h, width_2h, height_2h, c_2h, c_h;
    float *F_2h, *U_2h,  *R_h, *FR_2h;
    char *GT_2h;
    width_2h = grid_2h->width;
    height_2h = grid_2h->height;
    c_2h = 1+grid_2h->full_width;
    c_h = 1+grid_h->full_width;
    F_2h = grid_2h->F;
    R_h = grid_h->R;
    U_2h = grid_2h->U;
    GT_2h = grid_2h->GT;
    FR_2h = grid_2h->FR;
    for (jg_2h=0; jg_2h<height_2h; jg_2h++) {
        for (ig_2h=0; ig_2h<width_2h; ig_2h++) {
            F_2h[c_2h] = 4*GT_2h[c_2h]*R_h[c_h];
            U_2h[c_2h] = 0.0;
            c_2h +=1;
            c_h += 2;
        }
        c_2h += 2;
        c_h += 1+grid_h->full_width;
    }
    
}

void _injection_restriction_polar(grid_def *grid_h, grid_def *grid_2h) {
    int ig_2h, jg_2h, jg_2h_min, width_2h, height_2h, c_2h, c_h;
    float *F_2h, *U_2h,  *R_h, *FR_2h;
    char *GT_2h;
    width_2h = grid_2h->width;
    height_2h = grid_2h->height;
    F_2h = grid_2h->F;
    R_h = grid_h->R;
    U_2h = grid_2h->U;
    GT_2h = grid_2h->GT;
    FR_2h = grid_2h->FR;
    jg_2h_min = grid_2h->jg_min;
    c_2h = 1+grid_2h->full_width*(jg_2h_min+1);
    c_h = 1+grid_h->full_width*(2*jg_2h_min+1);
    for (jg_2h=jg_2h_min; jg_2h<height_2h; jg_2h++) {
        for (ig_2h=0; ig_2h<width_2h; ig_2h++) {
            F_2h[c_2h] = FR_2h[c_2h]*GT_2h[c_2h]*R_h[c_h];
            U_2h[c_2h] = 0.0;
            c_2h +=1;
            c_h += 2;
        }
        c_2h += 2;
        c_h += 2+grid_h->full_width;
    }
    
}



int poissonMG_multigrid_iterations(int nsteps, int nlevels, intArray niter_list, float w) {
    int s,l,n,k;
    grid_def *grid_h, *grid_2h;
    if (nlevels*2-1 != niter_list.dims[0]) return POISSON_BAD_ARRAY_DIM;
    if (solver.grid_levels < nlevels) return POISSON_ERROR_NOT_ENOUGH_GRIDS;
    n = solver.grid_levels;
    for (s=0; s<nsteps; s++) {
        k = 0;
        l = 0;
        for (l=0; l<nlevels-1; l++) {
            grid_h = grid_list[n-1-l];
            grid_2h = grid_list[n-1-l-1];
            solver.iterations_function(grid_h,niter_list.data[k],w);
            k++;
            solver.residu_function(grid_h);
            
            solver.restriction_function(grid_h, grid_2h);
        }
        grid_h = grid_list[n-nlevels];
        solver.iterations_function(grid_h,niter_list.data[k],w);
        k++;
        for (l=nlevels-2; l>=0; l--) {
            grid_h = grid_list[n-1-l];
            grid_2h = grid_list[n-1-l-1];
            solver.prolongation_function(grid_2h,grid_h);
            solver.iterations_function(grid_h,niter_list.data[k],w);
            k++;
        }
    }
    return 0;
}

int poissonMG_multigrid_Vcycles(int nsteps, int nlevels, int niter, int nu1, int nu2) {
    int s,l,n;
    grid_def *grid_h, *grid_2h;
    if (solver.grid_levels < nlevels) return POISSON_ERROR_NOT_ENOUGH_GRIDS;
    n = solver.grid_levels;
    for (s=0; s<nsteps; s++) {
        for (l=0; l<nlevels-1; l++) {
            grid_h = grid_list[n-1-l];
            grid_2h = grid_list[n-1-l-1];
            solver.iterations_function(grid_h,nu1,1.0);
            solver.residu_function(grid_h);
            solver.restriction_function(grid_h, grid_2h);
        }
        grid_h = grid_list[n-nlevels];
        solver.iterations_function(grid_h,niter,1.0);
        for (l=nlevels-2; l>=0; l--) {
            grid_h = grid_list[n-1-l];
            grid_2h = grid_list[n-1-l-1];
            solver.prolongation_function(grid_2h,grid_h);
            solver.iterations_function(grid_h,nu2,1.0);
        }
    }
    return 0;
}


int poissonMG_multigrid_full(int mcycle, int nlevels, int niter, int nu1, int nu2, float *wu) {
    grid_def *grid_h, *grid_2h;
    int n,k,level,l,c;
    if (solver.grid_levels < nlevels) return POISSON_ERROR_NOT_ENOUGH_GRIDS;
    n = solver.grid_levels;
    *wu = 0.0;
    for (c=0; c<mcycle; c++) {
     for (level=2; level<nlevels+1; level++) {
        l = 0;
        k = n-nlevels+l;
        solver.iterations_function(grid_list[k],niter,1.0);
        *wu += niter/pow(4,n-1-k);
        for (l=1; l<level; l++) {
            k = n-nlevels+l;
            grid_2h = grid_list[k-1];
            grid_h = grid_list[k];
            solver.prolongation_function(grid_2h,grid_h);
            solver.iterations_function(grid_h,nu2,1.0);
            *wu += nu2/pow(4,n-1-k);
        }
        for (l=level-1; l>=1; l--) {
            k = n-nlevels+l;
            grid_2h = grid_list[k-1];
            grid_h = grid_list[k];
            solver.residu_function(grid_h);
            solver.restriction_function(grid_h, grid_2h);
            solver.iterations_function(grid_2h,nu1,1.0);
            *wu += nu1/pow(4,n-1-(k-1));
        }
     }
    }
    l = 0;
    k = n-nlevels+l;
    solver.iterations_function(grid_list[k],niter,1.0);
    *wu += niter/pow(4,n-1-k);
    level = nlevels;
    for (l=1; l<level; l++) {
            k = n-nlevels+l;
            grid_2h = grid_list[k-1];
            grid_h = grid_list[k];
            solver.prolongation_function(grid_2h,grid_h);
            solver.iterations_function(grid_h,nu2,1.0);
            *wu += nu2/pow(4,n-1-k);
            
    }
    return 0;
}

/* openCL*/
#ifdef OPENCL

cl_uint numPlatforms;
cl_device_type deviceType;
cl_platform_id platform;
cl_device_id device;
size_t maxWorkGroupSize;
cl_uint maxWorkItemDimensions;
size_t maxWorkItemSizes[3];
cl_context context;
cl_program OpenCLProgram;
cl_command_queue commandQueue;
cl_kernel iterationPaire, iterationImpaire, iterationJacobiOne, iterationJacobiTwo, arrayNorm, iterationPaireColonne, iterationImpaireColonne, iterationPaireLigne, iterationImpaireLigne;

// noyaux openCL (ne traitent pas le cas polaire)
const char* programSource[] = {
"__kernel void iterationPaire(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned short j = get_global_id(1);",
    "unsigned int c = i+j*largeur;",
    "if ((((i&1)+(j&1))&1)==0) { // bit de parité de i+j\n",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationImpaire(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned short j = get_global_id(1);",
    "unsigned int c = i+j*largeur;",
    "if ((((i&1)+(j&1))&1)==1) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationImpaireColonne(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, short ligne, short colonne, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned int c = colonne+i*largeur;",
    "if ((i&1)==1) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationImpaireLigne(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, short ligne, short colonne, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned int c = i+ligne*largeur;",
    "if ((i&1)==1) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationPaireColonne(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, short ligne, short colonne, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned int c = colonne+i*largeur;",
    "if ((i&1)==0) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
    "if (i==1) {",
        "c = colonne+ligne*largeur;",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationPaireLigne(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, short ligne, short colonne,float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned int c = i+ligne*largeur;",
    "if ((i&1)==0) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void arrayNorm(__global float *U, __local float *buf, __global float *norm) {",
    "unsigned int i = get_local_id(0);",
    "unsigned int c = get_global_id(0);",
    "float u = U[c];",
    "buf[i] = u*u;",
    "barrier(CLK_LOCAL_MEM_FENCE);",
    "unsigned int j = get_local_size(0)/2;",
    "while (j!=0) {",
        "if (i < j) buf[i] += buf[i+j];",
        "barrier(CLK_LOCAL_MEM_FENCE);",
        "j /= 2;",
    "}",
    "if (i==0) norm[get_group_id(0)] = buf[0];",
"}"
};
#define SOURCE_NB_LINES 62
 
#define OPENCL_ERROR(e,msg) if (e<0) {printf("Erreur openCL : %s : %d\n",msg,e);return e;}

int poissonMG_opencl_platforms() {
    cl_int err;
    int p,k;
    char *name;
    size_t s;
    cl_platform_id* platforms;
    cl_uint numDev;
    cl_device_id* devices;
    err = clGetPlatformIDs(0,NULL,&numPlatforms);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    printf("Nombre de platformes openCL : %d\n",numPlatforms);
    platforms = (cl_platform_id*)malloc(sizeof(cl_platform_id)*numPlatforms);
    err = clGetPlatformIDs(numPlatforms,platforms,NULL);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    for (p=0; p<numPlatforms; p++) {
        err = clGetPlatformInfo(platforms[p],CL_PLATFORM_NAME,0,NULL,&s);
        OPENCL_ERROR(err,"clGetPlatformInfo")
        name = (char*)malloc(sizeof(char)*s);
        err = clGetPlatformInfo(platforms[p],CL_PLATFORM_NAME,s,name,NULL);
        OPENCL_ERROR(err,"clGetPlatformInfo")
        printf("Plateforme OpenCL %d : %s\n",p,name);
        free(name);
        //recherche des périphériques
        err = clGetDeviceIDs(platforms[p],CL_DEVICE_TYPE_ALL,0,NULL,&numDev);
        OPENCL_ERROR(err,"clGetDeviceIDs")
        printf("\t\tNombre de peripheriques : %d\n",numDev);
        devices = (cl_device_id*)malloc(sizeof(cl_device_id)*numDev);
        err = clGetDeviceIDs(platforms[p],CL_DEVICE_TYPE_ALL,numDev,devices,NULL);
        OPENCL_ERROR(err,"clGetDeviceIDs")
        for (k=0; k<numDev; k++) {
            err = clGetDeviceInfo(devices[k],CL_DEVICE_NAME,0,NULL,&s);
            OPENCL_ERROR(err,"clGetDeviceInfo")
            name = (char*)malloc(sizeof(char)*s);
            err = clGetDeviceInfo(devices[k],CL_DEVICE_NAME,s,name,NULL);
            OPENCL_ERROR(err,"clGetDeviceInfo")
            printf("\t\tPeripherique de calcul %d : %s\n",k,name);
            free(name); 
        }
        free(devices);
        printf("\n");
    }
    free(platforms);
    return 0;
}

int poissonMG_set_opencl_platform_device(int p, int d) {
    cl_int err;
    char *name;
    cl_platform_id* platforms;
    size_t s;
    cl_device_id* devices;
    cl_uint numDev;
    err = clGetPlatformIDs(0,NULL,&numPlatforms);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    platforms = (cl_platform_id*)malloc(sizeof(cl_platform_id)*numPlatforms);
    err = clGetPlatformIDs(numPlatforms,platforms,NULL);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    if ((p<0)||(p>=numPlatforms)) return POISSON_BAD_PLATFORM_NUMBER;
    platform = platforms[p];
    err = clGetPlatformInfo(platform,CL_PLATFORM_NAME,0,NULL,&s);
    OPENCL_ERROR(err,"clGetPlatformInfo")
    name = (char*)malloc(sizeof(char)*s);
    err = clGetPlatformInfo(platform,CL_PLATFORM_NAME,s,name,NULL);
    OPENCL_ERROR(err,"clGetPlatformInfo")
    printf("Plateforme OpenCL choisie : %s\n",name);
    free(name);
    free(platforms);
    err = clGetDeviceIDs(platform,CL_DEVICE_TYPE_ALL,0,NULL,&numDev);
    OPENCL_ERROR(err,"clGetDeviceIDs")
    devices = (cl_device_id*)malloc(sizeof(cl_device_id)*numDev);
    err = clGetDeviceIDs(platform,CL_DEVICE_TYPE_ALL,numDev,devices,NULL);
    if ((d<0)||(d>=numDev)) return POISSON_BAD_DEVICE_NUMBER;
    device = devices[d];
    free(devices);
    err = clGetDeviceInfo(device,CL_DEVICE_NAME,0,NULL,&s);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    name = (char*)malloc(sizeof(char)*s);
    err = clGetDeviceInfo(device,CL_DEVICE_NAME,s,name,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    printf("Peripherique de calcul choisi : %s",name);
    free(name);
    err = clGetDeviceInfo(device,CL_DEVICE_TYPE,sizeof(cl_device_type),&deviceType,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    if (deviceType==CL_DEVICE_TYPE_GPU) printf("\t (GPU)");
    else if (deviceType==CL_DEVICE_TYPE_CPU) printf("\t (CPU)");
    printf("\n");
    return 0;
}

int poissonMG_opencl_init() {
    int err;
    cl_context_properties context_props[3];
    
    if (platform==0) {
        err = poissonMG_set_opencl_platform_device(0,0);
        if (err<0) return err;
    }
    
    err = clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS,sizeof(maxWorkItemDimensions),&maxWorkItemDimensions,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    //printf("Max work group item dimensions : %d\n",maxWorkItemDimensions);
    err = clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_SIZES,sizeof(maxWorkItemSizes),maxWorkItemSizes,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    
    //for (d=0; d<maxWorkItemDimensions; d++) printf("Max work item in workgroup for dimension %d : %d\n",d+1,maxWorkItemSizes[d]);
    err = clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_GROUP_SIZE,sizeof(maxWorkGroupSize),&maxWorkGroupSize,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    //printf("Max work group size : %d\n",maxWorkGroupSize);
    context_props[0] = CL_CONTEXT_PLATFORM;
    context_props[1] = (cl_context_properties)platform;
    context_props[2] = 0;
    context = clCreateContextFromType(context_props,deviceType, NULL, NULL, &err);
    OpenCLProgram = clCreateProgramWithSource(context,SOURCE_NB_LINES,programSource,NULL,NULL);
    OPENCL_ERROR(err,"clCreateProgramWithSource")
    err = clBuildProgram(OpenCLProgram, 0, NULL, NULL, NULL, NULL);
    OPENCL_ERROR(err,"clBuildProgram")
    commandQueue = clCreateCommandQueue(context, device, 0, &err);
    OPENCL_ERROR(err,"clCreateCommandQueue") 
    iterationPaire = clCreateKernel(OpenCLProgram, "iterationPaire",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationImpaire = clCreateKernel(OpenCLProgram, "iterationImpaire",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    arrayNorm = clCreateKernel(OpenCLProgram,"arrayNorm",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationPaireColonne = clCreateKernel(OpenCLProgram,"iterationPaireColonne",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationImpaireColonne = clCreateKernel(OpenCLProgram,"iterationImpaireColonne",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationPaireLigne = clCreateKernel(OpenCLProgram,"iterationPaireLigne",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationImpaireLigne = clCreateKernel(OpenCLProgram,"iterationImpaireLigne",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    return 0;
}


int poissonMG_opencl_create_memory(grid_def* grid) {
    int err;
    // Buffers GPU
    grid->gpu_G = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*grid->full_size,grid->G,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    grid->gpu_D = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*grid->full_size,grid->D,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    grid->gpu_H = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*grid->full_size,grid->H,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    grid->gpu_B = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*grid->full_size,grid->B,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    grid->gpu_C = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*grid->full_size,grid->C,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    grid->gpu_F = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*grid->full_size,grid->F,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    grid->gpu_U = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR,sizeof(float)*grid->full_size,grid->U,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    grid->gpu_V = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR,sizeof(float)*grid->full_size,grid->U,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    return 0;
}

int poissonMG_opencl_create_memory_finegrid() {
    return poissonMG_opencl_create_memory(solver.fine_grid);   
}

int poissonMG_opencl_iterations(grid_def* grid, int niter, float w) {
    int err;
    size_t global_work_size[2];
    size_t local_work_size[2];
    size_t global_work_size_colonne[2];
    size_t local_work_size_colonne[2];
    size_t global_work_size_ligne[2];
    size_t local_work_size_ligne[2];
    int x,y,s;
    // offset pour ne pas commencer par les bords gauche et haut (rangées fictives à ne pas traiter)
    size_t global_work_offset[2] = {1,1};
    size_t global_work_offset_rangee[1] = {1};
    size_t wg_size, wg_size_mul;
    int n; 
    float w1;
    w1 = 1.0-w;
    
    // arguments des noyaux
    err = clSetKernelArg(iterationPaire, 0, sizeof(cl_mem),(void*)&(grid->gpu_D));
    err = clSetKernelArg(iterationPaire, 1, sizeof(cl_mem),(void*)&(grid->gpu_G));
    err = clSetKernelArg(iterationPaire, 2, sizeof(cl_mem),(void*)&(grid->gpu_H));
    err = clSetKernelArg(iterationPaire, 3, sizeof(cl_mem),(void*)&(grid->gpu_B));
    err = clSetKernelArg(iterationPaire, 4, sizeof(cl_mem),(void*)&(grid->gpu_C));
    err = clSetKernelArg(iterationPaire, 5, sizeof(cl_mem),(void*)&(grid->gpu_F));
    err = clSetKernelArg(iterationPaire, 6, sizeof(cl_mem),(void*)&(grid->gpu_U));
    err = clSetKernelArg(iterationPaire, 7, sizeof(short),&(grid->full_width));
    err = clSetKernelArg(iterationPaire, 8, sizeof(float),&w1);
    err = clSetKernelArg(iterationPaire, 9, sizeof(float),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationImpaire, 0, sizeof(cl_mem),(void*)&(grid->gpu_D));
    err = clSetKernelArg(iterationImpaire, 1, sizeof(cl_mem),(void*)&(grid->gpu_G));
    err = clSetKernelArg(iterationImpaire, 2, sizeof(cl_mem),(void*)&(grid->gpu_H));
    err = clSetKernelArg(iterationImpaire, 3, sizeof(cl_mem),(void*)&(grid->gpu_B));
    err = clSetKernelArg(iterationImpaire, 4, sizeof(cl_mem),(void*)&(grid->gpu_C));
    err = clSetKernelArg(iterationImpaire, 5, sizeof(cl_mem),(void*)&(grid->gpu_F));
    err = clSetKernelArg(iterationImpaire, 6, sizeof(cl_mem),(void*)&(grid->gpu_U));
    err = clSetKernelArg(iterationImpaire, 7, sizeof(grid->full_width),&(grid->full_width));
    err = clSetKernelArg(iterationImpaire, 8, sizeof(w1),&w1);
    err = clSetKernelArg(iterationImpaire, 9, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationPaireColonne, 0, sizeof(cl_mem),(void*)&(grid->gpu_D));
    err = clSetKernelArg(iterationPaireColonne, 1, sizeof(cl_mem),(void*)&(grid->gpu_G));
    err = clSetKernelArg(iterationPaireColonne, 2, sizeof(cl_mem),(void*)&(grid->gpu_H));
    err = clSetKernelArg(iterationPaireColonne, 3, sizeof(cl_mem),(void*)&(grid->gpu_B));
    err = clSetKernelArg(iterationPaireColonne, 4, sizeof(cl_mem),(void*)&(grid->gpu_C));
    err = clSetKernelArg(iterationPaireColonne, 5, sizeof(cl_mem),(void*)&(grid->gpu_F));
    err = clSetKernelArg(iterationPaireColonne, 6, sizeof(cl_mem),(void*)&(grid->gpu_U));
    err = clSetKernelArg(iterationPaireColonne, 7, sizeof(grid->full_width),&(grid->full_width));
    err = clSetKernelArg(iterationPaireColonne, 8, sizeof(grid->height),&(grid->height));
    err = clSetKernelArg(iterationPaireColonne, 9, sizeof(grid->width),&(grid->width));
    err = clSetKernelArg(iterationPaireColonne, 10, sizeof(w1),&w1);
    err = clSetKernelArg(iterationPaireColonne, 11, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationImpaireColonne, 0, sizeof(cl_mem),(void*)&(grid->gpu_D));
    err = clSetKernelArg(iterationImpaireColonne, 1, sizeof(cl_mem),(void*)&(grid->gpu_G));
    err = clSetKernelArg(iterationImpaireColonne, 2, sizeof(cl_mem),(void*)&(grid->gpu_H));
    err = clSetKernelArg(iterationImpaireColonne, 3, sizeof(cl_mem),(void*)&(grid->gpu_B));
    err = clSetKernelArg(iterationImpaireColonne, 4, sizeof(cl_mem),(void*)&(grid->gpu_C));
    err = clSetKernelArg(iterationImpaireColonne, 5, sizeof(cl_mem),(void*)&(grid->gpu_F));
    err = clSetKernelArg(iterationImpaireColonne, 6, sizeof(cl_mem),(void*)&(grid->gpu_U));
    err = clSetKernelArg(iterationImpaireColonne, 7, sizeof(grid->full_width),&(grid->full_width));
    err = clSetKernelArg(iterationImpaireColonne, 8, sizeof(grid->height),&(grid->height));
    err = clSetKernelArg(iterationImpaireColonne, 9, sizeof(grid->width),&(grid->width));
    err = clSetKernelArg(iterationImpaireColonne, 10, sizeof(w1),&w1);
    err = clSetKernelArg(iterationImpaireColonne, 11, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationPaireLigne, 0, sizeof(cl_mem),(void*)&(grid->gpu_D));
    err = clSetKernelArg(iterationPaireLigne, 1, sizeof(cl_mem),(void*)&(grid->gpu_G));
    err = clSetKernelArg(iterationPaireLigne, 2, sizeof(cl_mem),(void*)&(grid->gpu_H));
    err = clSetKernelArg(iterationPaireLigne, 3, sizeof(cl_mem),(void*)&(grid->gpu_B));
    err = clSetKernelArg(iterationPaireLigne, 4, sizeof(cl_mem),(void*)&(grid->gpu_C));
    err = clSetKernelArg(iterationPaireLigne, 5, sizeof(cl_mem),(void*)&(grid->gpu_F));
    err = clSetKernelArg(iterationPaireLigne, 6, sizeof(cl_mem),(void*)&(grid->gpu_U));
    err = clSetKernelArg(iterationPaireLigne, 7, sizeof(grid->full_width),&(grid->full_width));
    err = clSetKernelArg(iterationPaireLigne, 8, sizeof(grid->height),&(grid->height));
    err = clSetKernelArg(iterationPaireLigne, 9, sizeof(grid->width),&(grid->width));
    err = clSetKernelArg(iterationPaireLigne, 10, sizeof(w1),&w1);
    err = clSetKernelArg(iterationPaireLigne, 11, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationImpaireLigne, 0, sizeof(cl_mem),(void*)&(grid->gpu_D));
    err = clSetKernelArg(iterationImpaireLigne, 1, sizeof(cl_mem),(void*)&(grid->gpu_G));
    err = clSetKernelArg(iterationImpaireLigne, 2, sizeof(cl_mem),(void*)&(grid->gpu_H));
    err = clSetKernelArg(iterationImpaireLigne, 3, sizeof(cl_mem),(void*)&(grid->gpu_B));
    err = clSetKernelArg(iterationImpaireLigne, 4, sizeof(cl_mem),(void*)&(grid->gpu_C));
    err = clSetKernelArg(iterationImpaireLigne, 5, sizeof(cl_mem),(void*)&(grid->gpu_F));
    err = clSetKernelArg(iterationImpaireLigne, 6, sizeof(cl_mem),(void*)&(grid->gpu_U));
    err = clSetKernelArg(iterationImpaireLigne, 7, sizeof(grid->full_width),&(grid->full_width));
    err = clSetKernelArg(iterationImpaireLigne, 8, sizeof(grid->height),&(grid->height));
    err = clSetKernelArg(iterationImpaireLigne, 9, sizeof(grid->width),&(grid->width));
    err = clSetKernelArg(iterationImpaireLigne, 10, sizeof(w1),&w1);
    err = clSetKernelArg(iterationImpaireLigne, 11, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    
    // espace de travail : indices pour accéder aux tableaux (sans les bords)
    
    // Pour les noyaux iterationPaire et iterationImpaire, on traite un tableau dont les dimensions sont puissances de 2
    // on traite donc tout le tableau sauf la dernière colonne et la dernière ligne
    global_work_size[0] = grid->width-1;
    global_work_size[1] = grid->height-1;
    
    // recherche de la taille maximale pour les groupes de travail (dimensions puissances de 2)
    x = grid->width-1; y = grid->height-1;
    s = x*y;
    err = clGetKernelWorkGroupInfo(iterationPaire,device,CL_KERNEL_WORK_GROUP_SIZE,sizeof(wg_size),&wg_size,NULL);
    OPENCL_ERROR(err,"clGetKernelWorkGroupInfo");
    err = clGetKernelWorkGroupInfo(iterationPaire,device,CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE,sizeof(wg_size_mul),&wg_size,NULL);
    OPENCL_ERROR(err,"clGetKernelWorkGroupInfo");
    //printf("work group size : %d\n",wg_size);
    //printf("work group size multiple : %d\n",wg_size_mul);
    while (s > maxWorkGroupSize) { x /= 2; y /= 2; s = x*y;}
    local_work_size[0] = x;
    local_work_size[1] = y;
    //printf("group size : %d\n",s);
    
    //dernière colonne et dernière ligne
    global_work_size_colonne[0] = grid->height-1;
    y = grid->height-1;
    while (y > maxWorkGroupSize) {y /= 2;}
    local_work_size_colonne[0] = y;
    global_work_size_ligne[0] = grid->width-1;
    x = grid->width-1;
    while (x > maxWorkGroupSize) {x /= 2;}
    local_work_size_ligne[0] = x;
    
    // boucle d'itération  
    for (n=0; n<niter; n++) {
        err = clEnqueueNDRangeKernel(commandQueue,iterationPaire,2,global_work_offset,global_work_size,local_work_size,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_1")
        err = clEnqueueNDRangeKernel(commandQueue,iterationPaireLigne,1,global_work_offset_rangee,global_work_size_ligne,local_work_size_ligne,0,NULL,NULL); // dernière ligne
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_2")
        err = clEnqueueNDRangeKernel(commandQueue,iterationPaireColonne,1,global_work_offset_rangee,global_work_size_colonne,local_work_size_colonne,0,NULL,NULL); // dernière colonne
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_3")
        err = clEnqueueBarrier(commandQueue);
        OPENCL_ERROR(err,"clEnqueueBarrier");
        err = clEnqueueNDRangeKernel(commandQueue,iterationImpaire,2,global_work_offset,global_work_size,local_work_size,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_4")
        err = clEnqueueNDRangeKernel(commandQueue,iterationImpaireLigne,1,global_work_offset_rangee,global_work_size_ligne,local_work_size_ligne,0,NULL,NULL); // dernière ligne
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_5")
        err = clEnqueueNDRangeKernel(commandQueue,iterationImpaireColonne,1,global_work_offset_rangee,global_work_size_colonne,local_work_size_colonne,0,NULL,NULL); // dernière colonne
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel_6")
        err = clEnqueueBarrier(commandQueue);
        OPENCL_ERROR(err,"clEnqueueBarrier")
    }
    // récupération du tableau U
    err = clEnqueueReadBuffer(commandQueue, grid->gpu_U, CL_TRUE, 0,grid->full_size*sizeof(float), grid->U, 0, NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueReadBuffer")
    return 0;
}

int poissonMG_opencl_iterations_finegrid(int niter, float w) {
    return poissonMG_opencl_iterations(solver.fine_grid,niter,w);   
}

int poissonMG_opencl_get_array_norm(grid_def* grid, float *value) {
    int err;
    int workGroupSize;
    int numberOfGroups;
    short w,h,s;
    size_t global_work_size[1];
    size_t local_work_size[1];
    size_t src_origin[3] = {sizeof(float),1,0};
    size_t dst_origin[3] = {0,0,0};
    size_t region[3] = {(grid->width-1)*sizeof(float),grid->height-1,1};
    size_t src_row_pitch = grid->full_width*sizeof(float);
    size_t src_slice_pitch = 0;
    size_t dst_row_pitch = (grid->width-1)*sizeof(float);
    size_t dst_slice_pitch = 0;
    int k; 
    float n = 0.0;
    float *norm;
    cl_mem gpu_U1;
    cl_mem gpu_norm;
    
    w = grid->width-1; //largeur puissance de 2, on ignore la dernière colonne dans le calcul de la norme
    h = grid->height-1; // on ignore la dernière ligne
    s=w*h;
    if (maxWorkGroupSize < s) {
        workGroupSize = maxWorkGroupSize;
        numberOfGroups = s / workGroupSize;
    }
    else {
        workGroupSize = s;
        numberOfGroups = 1;
    }
    gpu_U1 = clCreateBuffer(context,CL_MEM_READ_WRITE,sizeof(float)*s,NULL,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    gpu_norm = clCreateBuffer(context,CL_MEM_WRITE_ONLY,sizeof(float)*numberOfGroups,NULL,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    err = clSetKernelArg(arrayNorm, 0, sizeof(cl_mem),(void*)&gpu_U1);
    err = clSetKernelArg(arrayNorm,1,sizeof(float)*workGroupSize,NULL);
    err = clSetKernelArg(arrayNorm,2,sizeof(cl_mem),(void*)&gpu_norm);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clEnqueueCopyBufferRect(commandQueue,grid->gpu_U,gpu_U1,src_origin,dst_origin,region,src_row_pitch,src_slice_pitch,dst_row_pitch,dst_slice_pitch,0,NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueCopyBufferRect")
    err = clEnqueueBarrier(commandQueue);
    OPENCL_ERROR(err,"clEnqueueBarrier");
    global_work_size[0] = grid->size;
    local_work_size[0] = workGroupSize;    
    err = clEnqueueNDRangeKernel(commandQueue,arrayNorm,1,NULL,global_work_size,local_work_size,0,NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueNDRangeKernel")
    err = clEnqueueBarrier(commandQueue);
    OPENCL_ERROR(err,"clEnqueueBarrier");
    norm = (float*)malloc(sizeof(float)*numberOfGroups);
    err = clEnqueueReadBuffer(commandQueue, gpu_norm, CL_TRUE, 0,numberOfGroups*sizeof(float), norm, 0, NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueReadBuffer");
    
    for (k=0; k<numberOfGroups; k++) n+= norm[k];
    *value = n/s;
    free(norm);
    clReleaseMemObject(gpu_U1);
    clReleaseMemObject(gpu_norm);
    return 0;
}

int poissonMG_opencl_get_array_norm_finegrid(float *value) {
       return poissonMG_opencl_get_array_norm(solver.fine_grid,value);
}

int poissonMG_opencl_release_memory(grid_def* grid) {
    clReleaseMemObject(grid->gpu_G);
    clReleaseMemObject(grid->gpu_D);
    clReleaseMemObject(grid->gpu_H);
    clReleaseMemObject(grid->gpu_B);
    clReleaseMemObject(grid->gpu_C);
    clReleaseMemObject(grid->gpu_F);
    clReleaseMemObject(grid->gpu_U);
    clReleaseMemObject(grid->gpu_V);
    return 0;
}

int poissonMG_opencl_release_memory_finegrid() {
    return poissonMG_opencl_release_memory(solver.fine_grid);
}

int poissonMG_opencl_release() {
    clReleaseKernel(iterationPaire);
    clReleaseKernel(iterationImpaire);
    clReleaseProgram(OpenCLProgram);
    clReleaseCommandQueue(commandQueue);
    clReleaseContext(context);
    return 0;
}
#endif


/*
    Création des grilles
    pw_min : exposant de 2 pour le nombre de mailles en largeur (grille 0)
    ph_min : exposant de 2 pour le nombre de mailles en hauteur
    levels : nombre de niveau de grilles
    w : taille réelle de la largeur (pour le calcul du pas d'espace)
    h : taille réelle de la hauteur

*/
int poissonMG_open(char type, int pw_min, int ph_min, int levels, float w, float h, char restriction) {
    int g,c;
    grid_def* grid;
    solver.type = type;
    solver.pw_min = pw_min;
    solver.ph_min = ph_min;
    solver.grid_levels = levels;
    solver.w = w;
    if (type==TYPE_POLAR) solver.w = 6.2831853;
    solver.h = h;
    grid_list = (grid_def**)malloc(sizeof(grid_def*)*levels);
    
    for (g=0; g<levels; g++) {
        grid = grid_list[g] = malloc(sizeof(grid_def));
        grid->index = g;
        grid->pw = pw_min+g;
        grid->ph = ph_min+g;
        grid->g = g;
        grid->r = (int)pow(2,g);
        grid->width = (int)pow(2,grid->pw)+1;
        grid->height = (int)pow(2,grid->ph)+1;
        if (type==TYPE_POLAR) grid->width -=1;
        grid->full_width = grid->width + 2;
        grid->full_height = grid->height +2;
        
        grid->size = (grid->width)*(grid->height);
        grid->full_size = (grid->full_width)*(grid->full_height);
        if (type==TYPE_POLAR) grid->dx = solver.w/(grid->width);
        else grid->dx = solver.w/(grid->width-1);
        grid->dy = solver.h/(grid->height-1);
        printf("dx = %g, dy = %g\n",grid->dx,grid->dy);
        grid->jg_min = 0;
        grid->G = (float*)malloc(sizeof(float)*grid->full_size);
        grid->D = (float*)malloc(sizeof(float)*grid->full_size);
        grid->H = (float*)malloc(sizeof(float)*grid->full_size);
        grid->B = (float*)malloc(sizeof(float)*grid->full_size);
        grid->C = (float*)malloc(sizeof(float)*grid->full_size);
        grid->F = (float*)malloc(sizeof(float)*grid->full_size);
        grid->U = (float*)malloc(sizeof(float)*grid->full_size);
        grid->R = (float*)malloc(sizeof(float)*grid->full_size);
        grid->FR = (float*)malloc(sizeof(float)*grid->full_size);
        grid->GT = (char*)malloc(sizeof(char)*grid->full_size);
        grid->U1 = (float*)malloc(sizeof(float)*grid->width*(2*grid->height-1));
        grid->P = (char*)malloc(sizeof(char)*grid->size);
        grid->M = (unsigned char*)malloc(sizeof(unsigned char)*grid->width*(2*grid->height-1));
        for (c=0; c<grid->full_size; c++) {
            grid->C[c] = 0.0; 
            grid->F[c] = 0.0;
            grid->D[c] = 0.0;
            grid->G[c] = 0.0;
            grid->H[c] = 0.0;
            grid->B[c] = 0.0;
            grid->U[c] = 0.0; 
            grid->R[c] = 0.0;
            grid->FR[c] = 0.0; 
            grid->GT[c] = 1;
        }
        for (c=0; c<grid->size; c++) grid->P[c] = 0;
    }
    solver.fine_grid = grid_list[levels-1];
    switch(type) {
    case TYPE_CARTESIAN:
        solver.source_function = _source;
        solver.laplacien_function = _laplacien;
        solver.neumann_function = _neumann;
        solver.interface_function = _interface;
        solver.prolongation_function = _linear_prolongation;
        switch(restriction) {
            case RESTRICTION_INJECTION : solver.restriction_function = _injection_restriction;
            case RESTRICTION_HALFWEIGHTING : solver.restriction_function = _halfweighting_restriction;
            case RESTRICTION_FULLWEIGHTING : solver.restriction_function = _fullweighting_restriction;
        }
        solver.iterations_function = _iterations;
        solver.residu_function = _residu;
        break;
    case TYPE_AXIAL:
        solver.source_function = _source_axial;
        solver.laplacien_function = _laplacien_axial;
        solver.neumann_function = _neumann_axial;
        solver.interface_function = _interface_axial;
        solver.prolongation_function = _linear_prolongation;
        switch(restriction) {
            case RESTRICTION_INJECTION : solver.restriction_function = _injection_restriction;
            case RESTRICTION_HALFWEIGHTING : solver.restriction_function = _halfweighting_restriction;
            case RESTRICTION_FULLWEIGHTING : solver.restriction_function = _fullweighting_restriction;
        }
        solver.iterations_function = _iterations;
        solver.residu_function = _residu;
        break;
    case TYPE_AXIAL_VECT:
        solver.source_function = _source;
        solver.laplacien_function = _laplacien_axial_vect;
        solver.neumann_function = _neumann_axial_vect;
        solver.interface_function = _interface_axial_vect;
        solver.prolongation_function = _linear_prolongation;
        switch(restriction) {
            case RESTRICTION_INJECTION : solver.restriction_function = _injection_restriction;
            case RESTRICTION_HALFWEIGHTING : solver.restriction_function = _halfweighting_restriction;
            case RESTRICTION_FULLWEIGHTING : solver.restriction_function = _fullweighting_restriction;
        }
        solver.iterations_function = _iterations;
        solver.residu_function = _residu;
        break;
    case TYPE_POLAR:
        solver.source_function = _source_polar;
        solver.laplacien_function = _laplacien_polar;
        solver.neumann_function = _neumann_polar;
        solver.interface_function = _interface_polar;
        solver.prolongation_function = _linear_prolongation_polar;
        solver.restriction_function = _injection_restriction_polar;
        solver.iterations_function = _iterations_polar;
        solver.residu_function = _residu_polar;
        break;
    }
    return 0;
}

int poissonMG_close(void) {
    int g;
    grid_def* grid;
    for (g=0; g<solver.grid_levels; g++) {
        grid = grid_list[g];
        free(grid->G);
        free(grid->D);
        free(grid->H);
        free(grid->B);
        free(grid->C);
        free(grid->F);
        free(grid->U);
        free(grid->R);
        free(grid->U1);
        free(grid->P);
        free(grid->M);
        free(grid->FR);
        free(grid);
    }
    free(grid_list);
    return 0;
}

