#include <Python.h> 
#include <numpy/npy_common.h>
#include <numpy/ndarrayobject.h>
#include "poisson.h"

#define POISSON_ERROR(e) {\
    if (e<0) {\
        switch(e) {\
        case POISSON_KERNEL_FILE_NOT_FOUND : PyErr_SetString(PyExc_ValueError,"kernel file not found"); return NULL;\
        case POISSON_INDEX_OUT_OF_RANGE : PyErr_SetString(PyExc_ValueError,"array index out of range"); return NULL;\
        case POISSON_BAD_ARRAY_DIM : PyErr_SetString(PyExc_ValueError,"bad array dimension"); return NULL;\
        case POISSON_BAD_INDEX_SPEC : PyErr_SetString(PyExc_ValueError,"bad index spec"); return NULL;\
        case POISSON_BAD_PLATFORM_NUMBER : PyErr_SetString(PyExc_ValueError,"bad opencl platform number"); return NULL;\
        }\
    PyErr_SetString(PyExc_ValueError,"Erreur"); return NULL;}\
}
#define ARG_ERROR(e) if (e) {PyErr_SetString(PyExc_ValueError,"bad arguments"); return NULL;}

static PyObject* py_poisson_open(PyObject *self, PyObject *args) {
    int pw,ph;
    ARG_ERROR(!PyArg_ParseTuple(args,"ii",&pw,&ph))
    POISSON_ERROR(poisson_open(pw,ph))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_close(PyObject *self, PyObject *args) {
    POISSON_ERROR(poisson_close())
    Py_RETURN_NONE;  
}

static PyObject* py_poisson_laplacien(PyObject *self, PyObject *args) {
    float dx,dy;
    ARG_ERROR(!PyArg_ParseTuple(args,"ff",&dx,&dy))
    POISSON_ERROR(poisson_laplacien(dx,dy))
    Py_RETURN_NONE;  
}

static PyObject* py_poisson_laplacien_axial(PyObject *self, PyObject *args) {
    float dz,dr;
    ARG_ERROR(!PyArg_ParseTuple(args,"ff",&dz,&dr))
    POISSON_ERROR(poisson_laplacien_axial(dz,dr))
    Py_RETURN_NONE;  
}

static PyObject* py_poisson_laplacien_axial_vect(PyObject *self, PyObject *args) {
    float dz,dr;
    ARG_ERROR(!PyArg_ParseTuple(args,"ff",&dz,&dr))
    POISSON_ERROR(poisson_laplacien_axial_vect(dz,dr))
    Py_RETURN_NONE;  
}

static PyObject* py_poisson_dirichlet_borders(PyObject *self, PyObject *args) {
    float value;
    ARG_ERROR(!PyArg_ParseTuple(args,"f",&value))
    POISSON_ERROR(poisson_dirichlet_borders(value))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_dirichlet_borders_axial(PyObject *self, PyObject *args) {
    float value;
    ARG_ERROR(!PyArg_ParseTuple(args,"f",&value))
    POISSON_ERROR(poisson_dirichlet_borders_axial(value))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_dirichlet_borders_axial_vect(PyObject *self, PyObject *args) {
    float value;
    ARG_ERROR(!PyArg_ParseTuple(args,"f",&value))
    POISSON_ERROR(poisson_dirichlet_borders_axial_vect(value))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_dirichlet(PyObject *self, PyObject *args) {
    int i,j;
    float value;
    ARG_ERROR(!PyArg_ParseTuple(args,"iif",&i,&j,&value))
    POISSON_ERROR(poisson_dirichlet(i,j,value))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_source(PyObject *self, PyObject *args) {
    int i,j;
    float s,dx,dy;
    ARG_ERROR(!PyArg_ParseTuple(args,"iifff",&i,&j,&s,&dx,&dy))
    POISSON_ERROR(poisson_source(i,j,s,dx,dy))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_source_axial(PyObject *self, PyObject *args) {
    int i,j;
    float s,dz,dr;
    ARG_ERROR(!PyArg_ParseTuple(args,"iifff",&i,&j,&s,&dz,&dr))
    POISSON_ERROR(poisson_source_axial(i,j,s,dz,dr))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_dirichlet_polygon(PyObject *self, PyObject *args) {
    PyObject *ilist_array,*jlist_array;
    float value;
    intArray ilist;
    intArray jlist;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"OOf",&ilist_array,&jlist_array,&value))
    ilist.ndims = jlist.ndims = 1;
    ilist.dims = (int*)malloc(sizeof(int));
    jlist.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(ilist_array);
    nj = PyList_Size(jlist_array);
    ilist.dims[0] = jlist.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;
    }
    ilist.data = (int*)malloc(sizeof(int)*ni);
    jlist.data = (int*)malloc(sizeof(int)*nj);
    for (k=0; k<ni; k++) {
        ilist.data[k] = (int)PyInt_AsLong(PyList_GetItem(ilist_array,k));
        jlist.data[k] = (int)PyInt_AsLong(PyList_GetItem(jlist_array,k));
    }
    POISSON_ERROR(poisson_dirichlet_polygon(ilist,jlist,value))
    free(ilist.data); free(jlist.data); free(ilist.dims); free(jlist.dims);
    Py_RETURN_NONE;
    
}

static PyObject* py_poisson_source_polygon(PyObject *self, PyObject *args) {
    PyObject *ilist_array,*jlist_array;
    float s,dx,dy;
    intArray ilist;
    intArray jlist;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"OOfff",&ilist_array,&jlist_array,&s,&dx,&dy))
    ilist.ndims = jlist.ndims = 1;
    ilist.dims = (int*)malloc(sizeof(int));
    jlist.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(ilist_array);
    nj = PyList_Size(jlist_array);
    ilist.dims[0] = jlist.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_source_polygon : incompatible list sizes");
        return NULL;
    }
    ilist.data = (int*)malloc(sizeof(int)*ni);
    jlist.data = (int*)malloc(sizeof(int)*nj);
    for (k=0; k<ni; k++) {
        ilist.data[k] = (int)PyInt_AsLong(PyList_GetItem(ilist_array,k));
        jlist.data[k] = (int)PyInt_AsLong(PyList_GetItem(jlist_array,k));
    }
    POISSON_ERROR(poisson_source_polygon(ilist,jlist,s,dx,dy))
    free(ilist.data); free(jlist.data); free(ilist.dims); free(jlist.dims);
    Py_RETURN_NONE;
    
}

static PyObject* py_poisson_source_polygon_axial(PyObject *self, PyObject *args) {
    PyObject *ilist_array,*jlist_array;
    float s,dz,dr;
    intArray ilist;
    intArray jlist;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"OOfff",&ilist_array,&jlist_array,&s,&dz,&dr))
    ilist.ndims = jlist.ndims = 1;
    ilist.dims = (int*)malloc(sizeof(int));
    jlist.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(ilist_array);
    nj = PyList_Size(jlist_array);
    ilist.dims[0] = jlist.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_source_polygon : incompatible list sizes");
        return NULL;
    }
    ilist.data = (int*)malloc(sizeof(int)*ni);
    jlist.data = (int*)malloc(sizeof(int)*nj);
    for (k=0; k<ni; k++) {
        ilist.data[k] = (int)PyInt_AsLong(PyList_GetItem(ilist_array,k));
        jlist.data[k] = (int)PyInt_AsLong(PyList_GetItem(jlist_array,k));
    }
    POISSON_ERROR(poisson_source_polygon_axial(ilist,jlist,s,dz,dr))
    free(ilist.data); free(jlist.data); free(ilist.dims); free(jlist.dims);
    Py_RETURN_NONE;
    
}

static PyObject* py_poisson_neumann(PyObject *self, PyObject *args) {
    int ip,jp,i,j,in,jn;
    float source,derivX,derivY,dx,dy;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiiiiifffff",&ip,&jp,&i,&j,&in,&jn,&source,&derivX,&derivY,&dx,&dy))
    POISSON_ERROR(poisson_neumann(ip,jp,i,j,in,jn,source,derivX,derivY,dx,dy))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_neumann_axial(PyObject *self, PyObject *args) {
    int ip,jp,i,j,in,jn;
    float source,derivX,derivY,dx,dy;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiiiiifffff",&ip,&jp,&i,&j,&in,&jn,&source,&derivX,&derivY,&dx,&dy))
    POISSON_ERROR(poisson_neumann_axial(ip,jp,i,j,in,jn,source,derivX,derivY,dx,dy))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_neumann_axial_vect(PyObject *self, PyObject *args) {
    int ip,jp,i,j,in,jn;
    float source,derivX,derivY,dx,dy;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiiiiifffff",&ip,&jp,&i,&j,&in,&jn,&source,&derivX,&derivY,&dx,&dy))
    POISSON_ERROR(poisson_neumann_axial_vect(ip,jp,i,j,in,jn,source,derivX,derivY,dx,dy))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_neumann_polygon(PyObject *self, PyObject *args) {
    PyObject *ilist_array,*jlist_array;
    float source,derivX,derivY,dx,dy;
    intArray ilist;
    intArray jlist;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"OOfffff",&ilist_array,&jlist_array,&source,&derivX,&derivY,&dx,&dy))
    ilist.ndims = jlist.ndims = 1;
    ilist.dims = (int*)malloc(sizeof(int));
    jlist.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(ilist_array);
    nj = PyList_Size(jlist_array);
    ilist.dims[0] = jlist.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;
    }
    ilist.data = (int*)malloc(sizeof(int)*ni);
    jlist.data = (int*)malloc(sizeof(int)*nj);
    for (k=0; k<ni; k++) {
        ilist.data[k] = (int)PyInt_AsLong(PyList_GetItem(ilist_array,k));
        jlist.data[k] = (int)PyInt_AsLong(PyList_GetItem(jlist_array,k));
    }
    POISSON_ERROR(poisson_neumann_polygon(ilist,jlist,source,derivX,derivY,dx,dy))
    free(ilist.data); free(jlist.data); free(ilist.dims); free(jlist.dims);
    Py_RETURN_NONE;
    
}

static PyObject* py_poisson_neumann_polygon_axial(PyObject *self, PyObject *args) {
    PyObject *ilist_array,*jlist_array;
    float source,derivX,derivY,dx,dy;
    intArray ilist;
    intArray jlist;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"OOfffff",&ilist_array,&jlist_array,&source,&derivX,&derivY,&dx,&dy))
    ilist.ndims = jlist.ndims = 1;
    ilist.dims = (int*)malloc(sizeof(int));
    jlist.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(ilist_array);
    nj = PyList_Size(jlist_array);
    ilist.dims[0] = jlist.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;
    }
    ilist.data = (int*)malloc(sizeof(int)*ni);
    jlist.data = (int*)malloc(sizeof(int)*nj);
    for (k=0; k<ni; k++) {
        ilist.data[k] = (int)PyInt_AsLong(PyList_GetItem(ilist_array,k));
        jlist.data[k] = (int)PyInt_AsLong(PyList_GetItem(jlist_array,k));
    }
    POISSON_ERROR(poisson_neumann_polygon_axial(ilist,jlist,source,derivX,derivY,dx,dy))
    free(ilist.data); free(jlist.data); free(ilist.dims); free(jlist.dims);
    Py_RETURN_NONE;
    
}

static PyObject* py_poisson_neumann_polygon_axial_vect(PyObject *self, PyObject *args) {
    PyObject *ilist_array,*jlist_array;
    float source,derivX,derivY,dx,dy;
    intArray ilist;
    intArray jlist;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"OOfffff",&ilist_array,&jlist_array,&source,&derivX,&derivY,&dx,&dy))
    ilist.ndims = jlist.ndims = 1;
    ilist.dims = (int*)malloc(sizeof(int));
    jlist.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(ilist_array);
    nj = PyList_Size(jlist_array);
    ilist.dims[0] = jlist.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;
    }
    ilist.data = (int*)malloc(sizeof(int)*ni);
    jlist.data = (int*)malloc(sizeof(int)*nj);
    for (k=0; k<ni; k++) {
        ilist.data[k] = (int)PyInt_AsLong(PyList_GetItem(ilist_array,k));
        jlist.data[k] = (int)PyInt_AsLong(PyList_GetItem(jlist_array,k));
    }
    POISSON_ERROR(poisson_neumann_polygon_axial_vect(ilist,jlist,source,derivX,derivY,dx,dy))
    free(ilist.data); free(jlist.data); free(ilist.dims); free(jlist.dims);
    Py_RETURN_NONE;
    
}

static PyObject* py_poisson_limit_polygon(PyObject *self, PyObject *args) {
    PyObject *ilist_array,*jlist_array,*limitType_array,*source_array,*value_array,*derivX_array,*derivY_array;
    float dx,dy;
    intArray ilist,jlist,limitType;
    floatArray source,value,derivX,derivY;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"OOOOOOOff",&ilist_array,&jlist_array,&limitType_array,&source_array,&value_array,&derivX_array,&derivY_array,&dx,&dy))
    ilist.ndims = jlist.ndims = limitType.ndims = source.ndims = value.ndims = derivX.ndims = derivY.ndims = 1;
    ilist.dims = (int*)malloc(sizeof(int));
    jlist.dims = (int*)malloc(sizeof(int));
    limitType.dims = (int*)malloc(sizeof(int));
    source.dims = (int*)malloc(sizeof(int));
    value.dims = (int*)malloc(sizeof(int));
    derivX.dims = (int*)malloc(sizeof(int));
    derivY.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(ilist_array);
    nj = PyList_Size(jlist_array);
    
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;
    }
    if ((ni!=PyList_Size(limitType_array))||(ni!=PyList_Size(source_array))||(ni!=PyList_Size(value_array))||(ni!=PyList_Size(derivX_array))||(ni!=PyList_Size(derivY_array))) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;   
    }
    ilist.dims[0] = jlist.dims[0] = limitType.dims[0] = source.dims[0] = value.dims[0] = derivX.dims[0] = derivY.dims[0] = ni;
    ilist.data = (int*)malloc(sizeof(int)*ni);
    jlist.data = (int*)malloc(sizeof(int)*ni);
    limitType.data = (int*)malloc(sizeof(int)*ni);
    source.data = (float*)malloc(sizeof(float)*ni);
    value.data = (float*)malloc(sizeof(float)*ni);
    derivX.data = (float*)malloc(sizeof(float)*ni);
    derivY.data = (float*)malloc(sizeof(float)*ni);
    for (k=0; k<ni; k++) {
        ilist.data[k] = (int)PyInt_AsLong(PyList_GetItem(ilist_array,k));
        jlist.data[k] = (int)PyInt_AsLong(PyList_GetItem(jlist_array,k));
        limitType.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(limitType_array,k));
        source.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(source_array,k));
        value.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(value_array,k));
        derivX.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(derivX_array,k));
        derivY.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(derivY_array,k));
    }
    POISSON_ERROR(poisson_limit_polygon(ilist,jlist,limitType,source,value,derivX,derivY,dx,dy))
    free(ilist.data); free(jlist.data); free(limitType.data); free(source.data); free(value.data); free(derivX.data); free(derivY.data); 
    free(ilist.dims); free(jlist.dims); free(limitType.dims); free(source.dims); free(value.dims); free(derivX.dims); free(derivY.dims);
    Py_RETURN_NONE;
    
}

static PyObject* py_poisson_limit_polygon_axial(PyObject *self, PyObject *args) {
    PyObject *ilist_array,*jlist_array,*limitType_array,*source_array,*value_array,*derivX_array,*derivY_array;
    float dx,dy;
    intArray ilist,jlist,limitType;
    floatArray source,value,derivX,derivY;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"OOOOOOOff",&ilist_array,&jlist_array,&limitType_array,&source_array,&value_array,&derivX_array,&derivY_array,&dx,&dy))
    ilist.ndims = jlist.ndims = limitType.ndims = source.ndims = value.ndims = derivX.ndims = derivY.ndims = 1;
    ilist.dims = (int*)malloc(sizeof(int));
    jlist.dims = (int*)malloc(sizeof(int));
    limitType.dims = (int*)malloc(sizeof(int));
    source.dims = (int*)malloc(sizeof(int));
    value.dims = (int*)malloc(sizeof(int));
    derivX.dims = (int*)malloc(sizeof(int));
    derivY.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(ilist_array);
    nj = PyList_Size(jlist_array);
    
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;
    }
    if ((ni!=PyList_Size(limitType_array))||(ni!=PyList_Size(source_array))||(ni!=PyList_Size(value_array))||(ni!=PyList_Size(derivX_array))||(ni!=PyList_Size(derivY_array))) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;   
    }
    ilist.dims[0] = jlist.dims[0] = limitType.dims[0] = source.dims[0] = value.dims[0] = derivX.dims[0] = derivY.dims[0] = ni;
    ilist.data = (int*)malloc(sizeof(int)*ni);
    jlist.data = (int*)malloc(sizeof(int)*ni);
    limitType.data = (int*)malloc(sizeof(int)*ni);
    source.data = (float*)malloc(sizeof(float)*ni);
    value.data = (float*)malloc(sizeof(float)*ni);
    derivX.data = (float*)malloc(sizeof(float)*ni);
    derivY.data = (float*)malloc(sizeof(float)*ni);
    for (k=0; k<ni; k++) {
        ilist.data[k] = (int)PyInt_AsLong(PyList_GetItem(ilist_array,k));
        jlist.data[k] = (int)PyInt_AsLong(PyList_GetItem(jlist_array,k));
        limitType.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(limitType_array,k));
        source.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(source_array,k));
        value.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(value_array,k));
        derivX.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(derivX_array,k));
        derivY.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(derivY_array,k));
    }
    POISSON_ERROR(poisson_limit_polygon_axial(ilist,jlist,limitType,source,value,derivX,derivY,dx,dy))
    free(ilist.data); free(jlist.data); free(limitType.data); free(source.data); free(value.data); free(derivX.data); free(derivY.data); 
    free(ilist.dims); free(jlist.dims); free(limitType.dims); free(source.dims); free(value.dims); free(derivX.dims); free(derivY.dims);
    Py_RETURN_NONE;
    
}

static PyObject* py_poisson_limit_polygon_axial_vect(PyObject *self, PyObject *args) {
    PyObject *ilist_array,*jlist_array,*limitType_array,*source_array,*value_array,*derivX_array,*derivY_array;
    float dx,dy;
    intArray ilist,jlist,limitType;
    floatArray source,value,derivX,derivY;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"OOOOOOOff",&ilist_array,&jlist_array,&limitType_array,&source_array,&value_array,&derivX_array,&derivY_array,&dx,&dy))
    ilist.ndims = jlist.ndims = limitType.ndims = source.ndims = value.ndims = derivX.ndims = derivY.ndims = 1;
    ilist.dims = (int*)malloc(sizeof(int));
    jlist.dims = (int*)malloc(sizeof(int));
    limitType.dims = (int*)malloc(sizeof(int));
    source.dims = (int*)malloc(sizeof(int));
    value.dims = (int*)malloc(sizeof(int));
    derivX.dims = (int*)malloc(sizeof(int));
    derivY.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(ilist_array);
    nj = PyList_Size(jlist_array);
    
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;
    }
    if ((ni!=PyList_Size(limitType_array))||(ni!=PyList_Size(source_array))||(ni!=PyList_Size(value_array))||(ni!=PyList_Size(derivX_array))||(ni!=PyList_Size(derivY_array))) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;   
    }
    ilist.dims[0] = jlist.dims[0] = limitType.dims[0] = source.dims[0] = value.dims[0] = derivX.dims[0] = derivY.dims[0] = ni;
    ilist.data = (int*)malloc(sizeof(int)*ni);
    jlist.data = (int*)malloc(sizeof(int)*ni);
    limitType.data = (int*)malloc(sizeof(int)*ni);
    source.data = (float*)malloc(sizeof(float)*ni);
    value.data = (float*)malloc(sizeof(float)*ni);
    derivX.data = (float*)malloc(sizeof(float)*ni);
    derivY.data = (float*)malloc(sizeof(float)*ni);
    for (k=0; k<ni; k++) {
        ilist.data[k] = (int)PyInt_AsLong(PyList_GetItem(ilist_array,k));
        jlist.data[k] = (int)PyInt_AsLong(PyList_GetItem(jlist_array,k));
        limitType.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(limitType_array,k));
        source.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(source_array,k));
        value.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(value_array,k));
        derivX.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(derivX_array,k));
        derivY.data[k] = (float)PyFloat_AsDouble(PyList_GetItem(derivY_array,k));
    }
    POISSON_ERROR(poisson_limit_polygon_axial_vect(ilist,jlist,limitType,source,value,derivX,derivY,dx,dy))
    free(ilist.data); free(jlist.data); free(limitType.data); free(source.data); free(value.data); free(derivX.data); free(derivY.data); 
    free(ilist.dims); free(jlist.dims); free(limitType.dims); free(source.dims); free(value.dims); free(derivX.dims); free(derivY.dims);
    Py_RETURN_NONE;
    
}


static PyObject* py_poisson_neumann_borders(PyObject *self, PyObject *args) {
    float source,derivX1,derivX2,derivY1,derivY2,dx,dy;
    ARG_ERROR(!PyArg_ParseTuple(args,"fffffff",&source,&derivX1,&derivX2,&derivY1,&derivY2,&dx,&dy))
    POISSON_ERROR(poisson_neumann_borders(source,derivX1,derivX2,derivY1,derivY2,dx,dy))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_neumann_borders_axial(PyObject *self, PyObject *args) {
    float source,derivX1,derivX2,derivY1,derivY2,dx,dy;
    ARG_ERROR(!PyArg_ParseTuple(args,"fffffff",&source,&derivX1,&derivX2,&derivY1,&derivY2,&dx,&dy))
    POISSON_ERROR(poisson_neumann_borders_axial(source,derivX1,derivX2,derivY1,derivY2,dx,dy))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_neumann_borders_axial_vect(PyObject *self, PyObject *args) {
    float source,derivX1,derivX2,derivY1,derivY2,dx,dy;
    ARG_ERROR(!PyArg_ParseTuple(args,"fffffff",&source,&derivX1,&derivX2,&derivY1,&derivY2,&dx,&dy))
    POISSON_ERROR(poisson_neumann_borders_axial_vect(source,derivX1,derivX2,derivY1,derivY2,dx,dy))
    Py_RETURN_NONE;
}


static PyObject* py_poisson_sinus_mode(PyObject *self, PyObject *args) {
    int mx,my;
    ARG_ERROR(!PyArg_ParseTuple(args,"ii",&mx,&my))
    POISSON_ERROR(poisson_sinus_mode(mx,my))
    Py_RETURN_NONE;
}


static PyObject* py_poisson_iterations(PyObject *self, PyObject *args) {
    int niter;
    float w;
    ARG_ERROR(!PyArg_ParseTuple(args,"if",&niter,&w))
    POISSON_ERROR(poisson_iterations(niter,w))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_iterations_2threads(PyObject *self, PyObject *args) {
    int niter;
    float w;
    ARG_ERROR(!PyArg_ParseTuple(args,"if",&niter,&w))
    POISSON_ERROR(poisson_iterations_2threads(niter,w))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_iterations_4threads(PyObject *self, PyObject *args) {
    int niter;
    float w;
    ARG_ERROR(!PyArg_ParseTuple(args,"if",&niter,&w))
    POISSON_ERROR(poisson_iterations_4threads(niter,w))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_get_array(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"i",&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poisson_get_array(&array,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}



static PyObject* py_poisson_get_derivX(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    float dx;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"fi",&dx,&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poisson_get_derivX(&array,dx,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}
static PyObject* py_poisson_get_derivY(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    float dy;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"fi",&dy,&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poisson_get_derivY(&array,dy,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}

static PyObject* py_poisson_get_derivRUR_axial(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    float dr;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"fi",&dr,&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poisson_get_derivRUR_axial(&array,dr,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}

static PyObject* py_poisson_get_array_norm(PyObject *self, PyObject *args) {
    float norm;
    POISSON_ERROR(poisson_get_array_norm(&norm))
    return Py_BuildValue("f",norm);
}

#ifdef OPENCL
static PyObject* py_poisson_opencl_platforms(PyObject *self, PyObject *args) {
    POISSON_ERROR(poisson_opencl_platforms())
    Py_RETURN_NONE;  
}

static PyObject* py_poisson_set_opencl_platform_device(PyObject *self, PyObject *args) {
    int p,d;
    ARG_ERROR(!PyArg_ParseTuple(args,"ii",&p,&d))
    POISSON_ERROR(poisson_set_opencl_platform_device(p,d))
    Py_RETURN_NONE;
}
static PyObject* py_poisson_opencl_init(PyObject *self, PyObject *args) {
    POISSON_ERROR(poisson_opencl_init())
    Py_RETURN_NONE;  
}

static PyObject* py_poisson_opencl_create_memory(PyObject *self, PyObject *args) {
    POISSON_ERROR(poisson_opencl_create_memory())
    Py_RETURN_NONE;  
}

static PyObject* py_poisson_opencl_release_memory(PyObject *self, PyObject *args) {
    POISSON_ERROR(poisson_opencl_release_memory())
    Py_RETURN_NONE;  
}

static PyObject* py_poisson_opencl_release(PyObject *self, PyObject *args) {
    POISSON_ERROR(poisson_opencl_release())
    Py_RETURN_NONE;  
}

static PyObject* py_poisson_opencl_iterations(PyObject *self, PyObject *args) {
    int niter;
    float w;
    ARG_ERROR(!PyArg_ParseTuple(args,"if",&niter,&w))
    POISSON_ERROR(poisson_opencl_iterations(niter,w))
    Py_RETURN_NONE;
}

static PyObject* py_poisson_opencl_get_array_norm(PyObject *self, PyObject *args) {
    float norm;
    POISSON_ERROR(poisson_opencl_get_array_norm(&norm))
    return Py_BuildValue("f",norm);
}
#endif


#ifdef OPENCL
static PyMethodDef pypoissonMethods[] = {
    {"open",py_poisson_open,METH_VARARGS},
    {"close",py_poisson_close,METH_NOARGS},
    {"laplacien",py_poisson_laplacien,METH_VARARGS},
    {"laplacien_axial",py_poisson_laplacien_axial,METH_VARARGS},
    {"laplacien_axial_vect",py_poisson_laplacien_axial_vect,METH_VARARGS},
    {"dirichlet_borders",py_poisson_dirichlet_borders,METH_VARARGS},
    {"dirichlet_borders_axial",py_poisson_dirichlet_borders_axial,METH_VARARGS},
    {"dirichlet_borders_axial_vect",py_poisson_dirichlet_borders_axial_vect,METH_VARARGS},
    {"dirichlet",py_poisson_dirichlet,METH_VARARGS},
    {"dirichlet_polygon",py_poisson_dirichlet_polygon,METH_VARARGS},
    {"source",py_poisson_source,METH_VARARGS},
    {"source_axial",py_poisson_source_axial,METH_VARARGS},
    {"source_polygon",py_poisson_source_polygon,METH_VARARGS},
    {"source_polygon_axial",py_poisson_source_polygon_axial,METH_VARARGS},
    {"neumann",py_poisson_neumann,METH_VARARGS},
    {"neumann_axial",py_poisson_neumann_axial,METH_VARARGS},
    {"neumann_axial_vect",py_poisson_neumann_axial_vect,METH_VARARGS},
    {"neumann_polygon",py_poisson_neumann_polygon,METH_VARARGS},
    {"neumann_polygon_axial",py_poisson_neumann_polygon_axial,METH_VARARGS},
    {"neumann_polygon_axial_vect",py_poisson_neumann_polygon_axial_vect,METH_VARARGS},
    {"neumann_borders",py_poisson_neumann_borders,METH_VARARGS},
    {"neumann_borders_axial",py_poisson_neumann_borders_axial,METH_VARARGS},
    {"neumann_borders_axial_vect",py_poisson_neumann_borders_axial_vect,METH_VARARGS},
    {"limit_polygon",py_poisson_limit_polygon,METH_VARARGS},
    {"limit_polygon_axial",py_poisson_limit_polygon_axial,METH_VARARGS},
    {"limit_polygon_axial_vect",py_poisson_limit_polygon_axial_vect,METH_VARARGS},
    {"sinus_mode",py_poisson_sinus_mode,METH_VARARGS}, 
    {"iterations",py_poisson_iterations,METH_VARARGS},
    {"iterations_2threads",py_poisson_iterations_2threads,METH_VARARGS},
    {"iterations_4threads",py_poisson_iterations_4threads,METH_VARARGS},
    {"get_array",py_poisson_get_array,METH_VARARGS},
    {"get_derivX",py_poisson_get_derivX,METH_VARARGS},
    {"get_derivY",py_poisson_get_derivY,METH_VARARGS},
    {"get_derivRUR_axial",py_poisson_get_derivRUR_axial,METH_VARARGS},
    {"get_array_norm",py_poisson_get_array_norm,METH_NOARGS},
    {"opencl_platforms",py_poisson_opencl_platforms,METH_NOARGS},
    {"set_opencl_platform_device",py_poisson_set_opencl_platform_device,METH_VARARGS},
    {"opencl_init",py_poisson_opencl_init,METH_NOARGS},
    {"opencl_create_memory",py_poisson_opencl_create_memory,METH_NOARGS},
    {"opencl_release_memory",py_poisson_opencl_release_memory,METH_NOARGS},
    {"opencl_iterations",py_poisson_opencl_iterations,METH_VARARGS},
    {"opencl_release",py_poisson_opencl_release,METH_NOARGS},
    {"opencl_get_array_norm",py_poisson_opencl_get_array_norm,METH_NOARGS},
    {NULL,NULL}
};
#else
static PyMethodDef pypoissonMethods[] = {
    {"open",py_poisson_open,METH_VARARGS},
    {"close",py_poisson_close,METH_NOARGS},
    {"laplacien",py_poisson_laplacien,METH_VARARGS},
    {"laplacien_axial",py_poisson_laplacien_axial,METH_VARARGS},
    {"laplacien_axial_vect",py_poisson_laplacien_axial_vect,METH_VARARGS},
    {"dirichlet_borders",py_poisson_dirichlet_borders,METH_VARARGS},
    {"dirichlet_borders_axial",py_poisson_dirichlet_borders_axial,METH_VARARGS},
    {"dirichlet_borders_axial_vect",py_poisson_dirichlet_borders_axial_vect,METH_VARARGS},
    {"dirichlet",py_poisson_dirichlet,METH_VARARGS},
    {"dirichlet_polygon",py_poisson_dirichlet_polygon,METH_VARARGS},
    {"neumann",py_poisson_neumann,METH_VARARGS},
    {"neumann_axial",py_poisson_neumann_axial,METH_VARARGS},
    {"neumann_axial_vect",py_poisson_neumann_axial_vect,METH_VARARGS},
    {"source",py_poisson_source,METH_VARARGS},
    {"source_axial",py_poisson_source_axial,METH_VARARGS},
    {"source_polygon",py_poisson_source_polygon,METH_VARARGS},
    {"source_polygon_axial",py_poisson_source_polygon_axial,METH_VARARGS},
    {"neumann_polygon_axial_vect",py_poisson_neumann_polygon_axial_vect,METH_VARARGS},
    {"neumann_polygon",py_poisson_neumann_polygon,METH_VARARGS},
    {"neumann_polygon_axial",py_poisson_neumann_polygon_axial,METH_VARARGS},
    {"neumann_borders",py_poisson_neumann_borders,METH_VARARGS},
    {"neumann_borders_axial",py_poisson_neumann_borders_axial,METH_VARARGS},
    {"neumann_borders_axial_vect",py_poisson_neumann_borders_axial_vect,METH_VARARGS},
    {"limit_polygon",py_poisson_limit_polygon,METH_VARARGS},
    {"limit_polygon_axial",py_poisson_limit_polygon_axial,METH_VARARGS},
    {"limit_polygon_axial_vect",py_poisson_limit_polygon_axial_vect,METH_VARARGS},
    {"sinus_mode",py_poisson_sinus_mode,METH_VARARGS}, 
    {"iterations",py_poisson_iterations,METH_VARARGS},
    {"iterations_2threads",py_poisson_iterations_2threads,METH_VARARGS},
    {"iterations_4threads",py_poisson_iterations_4threads,METH_VARARGS},
    {"get_array",py_poisson_get_array,METH_VARARGS},
    {"get_derivX",py_poisson_get_derivX,METH_VARARGS},
    {"get_derivY",py_poisson_get_derivY,METH_VARARGS},
    {"get_derivRUR_axial",py_poisson_get_derivRUR_axial,METH_VARARGS},
    {"get_array_norm",py_poisson_get_array_norm,METH_NOARGS},
    {NULL,NULL}
};
#endif

void initpypoisson(void) {
    Py_InitModule("pypoisson",pypoissonMethods);
    import_array();
}
