#include <Python.h> 
#include <numpy/npy_common.h>
#include <numpy/ndarrayobject.h>
#include "poissonMG.h"
 
#define POISSON_ERROR(e) {\
    if (e<0) {\
        switch(e) {\
        case POISSON_KERNEL_FILE_NOT_FOUND : PyErr_SetString(PyExc_ValueError,"kernel file not found"); return NULL;\
        case POISSON_INDEX_OUT_OF_RANGE : PyErr_SetString(PyExc_ValueError,"array index out of range"); return NULL;\
        case POISSON_BAD_ARRAY_DIM : PyErr_SetString(PyExc_ValueError,"bad array dimension"); return NULL;\
        case POISSON_BAD_INDEX_SPEC : PyErr_SetString(PyExc_ValueError,"bad index spec"); return NULL;\
        case POISSON_BAD_PLATFORM_NUMBER : PyErr_SetString(PyExc_ValueError,"bad opencl platform number"); return NULL;\
        case POISSON_ERROR_NOT_ENOUGH_GRIDS : PyErr_SetString(PyExc_ValueError,"not enough grids"); return NULL;\
        case POISSON_BAD_POLYGON_DEFINITION : PyErr_SetString(PyExc_ValueError,"bad polygon definition"); return NULL;\
        }\
    PyErr_SetString(PyExc_ValueError,"Erreur"); return NULL;}\
}
#define ARG_ERROR(e) if (e) {PyErr_SetString(PyExc_ValueError,"bad arguments"); return NULL;}
 
static PyObject* py_poissonMG_open(PyObject *self, PyObject *args) {
    int pw_min,ph_min,levels;
    float w,h;
    char type,restriction;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiiiffi",&type,&pw_min,&ph_min,&levels,&w,&h,&restriction))
    POISSON_ERROR(poissonMG_open(type,pw_min,ph_min,levels,w,h,restriction))
    Py_RETURN_NONE;
}
 
static PyObject* py_poissonMG_close(PyObject *self, PyObject *args) {
    POISSON_ERROR(poissonMG_close())
    Py_RETURN_NONE;  
}

static PyObject* py_poissonMG_laplacien(PyObject *self, PyObject *args) {
    POISSON_ERROR(poissonMG_laplacien())
    Py_RETURN_NONE;  
}

static PyObject* py_poissonMG_source(PyObject *self, PyObject *args) {
    int i,j;
    float s;
    ARG_ERROR(!PyArg_ParseTuple(args,"iif",&i,&j,&s))
    POISSON_ERROR(poissonMG_source(i,j,s))
    Py_RETURN_NONE;
} 

static PyObject* py_poissonMG_source_polygon(PyObject *self, PyObject *args) {
    PyObject *li_list_array, *lj_list_array;
    float value;
    int i_start,j_start;
    intArray li_list, lj_list;
    int ni,nj;
    int k; 
    ARG_ERROR(!PyArg_ParseTuple(args,"iiOOf",&i_start,&j_start,&li_list_array,&lj_list_array,&value))
    
    li_list.ndims = lj_list.ndims = 1;
    li_list.dims = (int*)malloc(sizeof(int));
    lj_list.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(li_list_array);
    nj = PyList_Size(lj_list_array);
    li_list.dims[0] = lj_list.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_source_polygon : incompatible list sizes");
        return NULL;
    }
    li_list.data = (int*)malloc(sizeof(int)*ni);
    lj_list.data = (int*)malloc(sizeof(int)*nj);
    for (k=0; k<ni; k++) {
        li_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(li_list_array,k));
        lj_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(lj_list_array,k));
    }
     
    
    POISSON_ERROR(poissonMG_source_polygon(i_start,j_start,li_list,lj_list,value))
    free(li_list.data); free(lj_list.data); 
    free(li_list.dims); free(lj_list.dims);
    Py_RETURN_NONE;    
}

static PyObject* py_poissonMG_source_rect(PyObject *self, PyObject *args) {
    int cx,cy,width,height;
    float s;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiiif",&cx,&cy,&width,&height,&s))
    POISSON_ERROR(poissonMG_source_rect(cx,cy,width,height,s))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_dirichlet(PyObject *self, PyObject *args) {
    int i,j;
    float value;
    ARG_ERROR(!PyArg_ParseTuple(args,"iif",&i,&j,&value))
    POISSON_ERROR(poissonMG_dirichlet(i,j,value))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_dirichlet_borders(PyObject *self, PyObject *args) {
    float value;
    ARG_ERROR(!PyArg_ParseTuple(args,"f",&value))
    POISSON_ERROR(poissonMG_dirichlet_borders(value))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_dirichlet_polygon(PyObject *self, PyObject *args) {
    PyObject *li_list_array, *lj_list_array;
    float value;
    int i_start,j_start;
    intArray li_list, lj_list;
    int ni,nj;
    int k; 
    ARG_ERROR(!PyArg_ParseTuple(args,"iiOOf",&i_start,&j_start,&li_list_array,&lj_list_array,&value))
    
    li_list.ndims = lj_list.ndims = 1;
    li_list.dims = (int*)malloc(sizeof(int));
    lj_list.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(li_list_array);
    nj = PyList_Size(lj_list_array);
    li_list.dims[0] = lj_list.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_dirichlet_polygon : incompatible list sizes");
        return NULL;
    }
    li_list.data = (int*)malloc(sizeof(int)*ni);
    lj_list.data = (int*)malloc(sizeof(int)*nj); 
    for (k=0; k<ni; k++) {
        li_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(li_list_array,k));
        lj_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(lj_list_array,k));
    }
    
    
    POISSON_ERROR(poissonMG_dirichlet_polygon(i_start,j_start,li_list,lj_list,value))
    free(li_list.data); free(lj_list.data); 
    free(li_list.dims); free(lj_list.dims);
    Py_RETURN_NONE;    
}

static PyObject* py_poissonMG_neumann_borders(PyObject *self, PyObject *args) {
    float source,derivX1,derivX2,derivY1,derivY2;
    ARG_ERROR(!PyArg_ParseTuple(args,"fffff",&source,&derivX1,&derivX2,&derivY1,&derivY2))
    POISSON_ERROR(poissonMG_neumann_borders(source,derivX1,derivX2,derivY1,derivY2))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_neumann_border_ext_polar(PyObject *self, PyObject *args) {
    float source,value;
    ARG_ERROR(!PyArg_ParseTuple(args,"ff",&source,&value))
    POISSON_ERROR(poissonMG_neumann_border_ext_polar(source,value))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_neumann_border_int_polar(PyObject *self, PyObject *args) {
    float source,value;
    int j;
    ARG_ERROR(!PyArg_ParseTuple(args,"iff",&j,&source,&value))
    POISSON_ERROR(poissonMG_neumann_border_int_polar(j,source,value))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_neumann_polygon(PyObject *self, PyObject *args) {
    int i_start,j_start;
    PyObject *li_list_array, *lj_list_array, *derivX_list_array, *derivY_list_array;
    float source;
    intArray li_list, lj_list;
    floatArray derivX_list, derivY_list;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiOOOOf",&i_start,&j_start,&li_list_array,&lj_list_array,&derivX_list_array,&derivY_list_array,&source))
    li_list.ndims = lj_list.ndims = 1;
    li_list.dims = (int*)malloc(sizeof(int));
    lj_list.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(li_list_array);
    nj = PyList_Size(lj_list_array);
    li_list.dims[0] = lj_list.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_neumann_polygon : incompatible list sizes");
        return NULL;
    }
    li_list.data = (int*)malloc(sizeof(int)*ni);
    lj_list.data = (int*)malloc(sizeof(int)*nj);
    for (k=0; k<ni; k++) {
        li_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(li_list_array,k));
        lj_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(lj_list_array,k));
    }
    derivX_list.ndims = derivY_list.ndims = 1;
    derivX_list.dims = (int*)malloc(sizeof(int));
    derivY_list.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(derivX_list_array);
    nj = PyList_Size(derivY_list_array);
    derivX_list.dims[0] = derivY_list.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_neumann_polygon : incompatible list sizes");
        return NULL;
    }
    derivX_list.data = (float*)malloc(sizeof(float)*ni);
    derivY_list.data = (float*)malloc(sizeof(float)*nj);
    for (k=0; k<ni; k++) {
        derivX_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(derivX_list_array,k));
        derivY_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(derivY_list_array,k));
    }
    
    POISSON_ERROR(poissonMG_neumann_polygon(i_start,j_start,li_list,lj_list,derivX_list,derivY_list,source))
    free(li_list.data); free(lj_list.data); free(li_list.dims); free(lj_list.dims);
    free(derivX_list.data); free(derivY_list.data); free(derivX_list.dims); free(derivY_list.dims);
    Py_RETURN_NONE;    
}

static PyObject* py_poissonMG_interface_polygon(PyObject *self, PyObject *args) {
    int i_start,j_start;
    PyObject *li_list_array, *lj_list_array;
    float source,a1,a2;
    intArray li_list, lj_list;
    int ni,nj;
    int k;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiOOfff",&i_start,&j_start,&li_list_array,&lj_list_array,&source,&a1,&a2))
    li_list.ndims = lj_list.ndims = 1;
    li_list.dims = (int*)malloc(sizeof(int));
    lj_list.dims = (int*)malloc(sizeof(int));
    ni = PyList_Size(li_list_array);
    nj = PyList_Size(lj_list_array);
    li_list.dims[0] = lj_list.dims[0] = ni;
    if (ni!=nj) {
        PyErr_SetString(PyExc_ValueError,"poisson_interface_polygon : incompatible list sizes");
        return NULL;
    }
    li_list.data = (int*)malloc(sizeof(int)*ni);
    lj_list.data = (int*)malloc(sizeof(int)*nj);
    for (k=0; k<ni; k++) {
        li_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(li_list_array,k));
        lj_list.data[k] = (int)PyLong_AsLong(PyList_GetItem(lj_list_array,k));
    }
    
    
    POISSON_ERROR(poissonMG_interface_polygon(i_start,j_start,li_list,lj_list,source,a1,a2))
    free(li_list.data); free(lj_list.data); free(li_list.dims); free(lj_list.dims);
    Py_RETURN_NONE;    
}



static PyObject* py_poissonMG_sinus_mode(PyObject *self, PyObject *args) {
    int mx,my;
    ARG_ERROR(!PyArg_ParseTuple(args,"ii",&mx,&my))
    POISSON_ERROR(poissonMG_sinus_mode(mx,my))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_step_mode(PyObject *self, PyObject *args) {
    POISSON_ERROR(poissonMG_step_mode())
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_finegrid_iterations(PyObject *self, PyObject *args) {
    int niter;
    float w;
    ARG_ERROR(!PyArg_ParseTuple(args,"if",&niter,&w))
    POISSON_ERROR(poissonMG_finegrid_iterations(niter,w))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_finegrid_iterations_2threads(PyObject *self, PyObject *args) {
    int niter;
    float w;
    ARG_ERROR(!PyArg_ParseTuple(args,"if",&niter,&w))
    POISSON_ERROR(poissonMG_finegrid_iterations_2threads(niter,w))
    Py_RETURN_NONE;
} 
 
static PyObject* py_poissonMG_finegrid_iterations_4threads(PyObject *self, PyObject *args) {
    int niter;
    float w;
    ARG_ERROR(!PyArg_ParseTuple(args,"if",&niter,&w))
    POISSON_ERROR(poissonMG_finegrid_iterations_4threads(niter,w))
    Py_RETURN_NONE;
}

 

static PyObject* py_poissonMG_multigrid_iterations(PyObject *self, PyObject *args) {
    int nsteps,nlevels,k,n;
    PyObject *niter_list_array;
    intArray niter_list;
    float w;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiOf",&nsteps,&nlevels,&niter_list_array,&w))
    niter_list.ndims = 1;
    niter_list.dims = (int*)malloc(sizeof(int));
    n = niter_list.dims[0] = PyList_Size(niter_list_array);
    niter_list.data = (int*)malloc(sizeof(int)*n);
    for (k=0; k<n; k++) niter_list.data[k]  =  (int)PyLong_AsLong(PyList_GetItem(niter_list_array,k));
    POISSON_ERROR(poissonMG_multigrid_iterations(nsteps,nlevels,niter_list,w))
    free(niter_list.data); free(niter_list.dims);
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_multigrid_full(PyObject *self, PyObject *args) {
    int mcycle,nlevels,niter,nu1,nu2;
    float wu;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiiii",&mcycle,&nlevels,&niter,&nu1,&nu2))
    POISSON_ERROR(poissonMG_multigrid_full(mcycle,nlevels,niter,nu1,nu2,&wu))
    return Py_BuildValue("f",wu);
}

static PyObject* py_poissonMG_multigrid_Vcycles(PyObject *self, PyObject *args) {
    int nsteps,nlevels,niter,nu1,nu2;
    ARG_ERROR(!PyArg_ParseTuple(args,"iiiii",&nsteps,&nlevels,&niter,&nu1,&nu2))
    POISSON_ERROR(poissonMG_multigrid_Vcycles(nsteps,nlevels,niter,nu1,nu2))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_get_array(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"i",&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poissonMG_get_array(&array,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}

static PyObject* py_poissonMG_get_mask(PyObject *self, PyObject *args) {
    byteArray array;
    PyObject *numpy_array;
    void *data;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"i",&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poissonMG_get_mask(&array,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_UINT8);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(unsigned char)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}

static PyObject* py_poissonMG_get_derivX(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"i",&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poissonMG_get_derivX(&array,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}

static PyObject* py_poissonMG_get_derivY(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"i",&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poissonMG_get_derivY(&array,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}

static PyObject* py_poissonMG_get_derivRUR_axial(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"i",&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poissonMG_get_derivRUR_axial(&array,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}

static PyObject* py_poissonMG_get_derivTheta_polar(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    npy_intp dims[2], c[2];
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poissonMG_get_derivTheta_polar(&array))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}

static PyObject* py_poissonMG_get_RUR_axial(PyObject *self, PyObject *args) {
    floatArray array;
    PyObject *numpy_array;
    void *data;
    int symetry;
    npy_intp dims[2], c[2];
    ARG_ERROR(!PyArg_ParseTuple(args,"i",&symetry))
    array.ndims = 2;
    array.dims = (int*)malloc(sizeof(int)*array.ndims);
    POISSON_ERROR(poissonMG_get_RUR_axial(&array,symetry))
    dims[0] = array.dims[0];
    dims[1] = array.dims[1];
    c[0] = 0; c[1] = 0;
    numpy_array = PyArray_SimpleNew(2,dims,NPY_FLOAT);
    data = PyArray_GetPtr((PyArrayObject*)numpy_array,c);
    memcpy(data,array.data,sizeof(float)*dims[0]*dims[1]);
    free(array.dims);
    return Py_BuildValue("O",numpy_array); 
}

static PyObject* py_poissonMG_get_finegrid_residu_norm(PyObject *self, PyObject *args) {
    float norm;
    POISSON_ERROR(poissonMG_get_finegrid_residu_norm(&norm))
    return Py_BuildValue("f",norm);
}

static PyObject* py_poissonMG_get_finegrid_array_norm(PyObject *self, PyObject *args) {
    float norm;
    POISSON_ERROR(poissonMG_get_finegrid_array_norm(&norm))
    return Py_BuildValue("f",norm);
}

static PyObject* py_poissonMG_init_array(PyObject *self, PyObject *args) {
    POISSON_ERROR(poissonMG_init_array())
    Py_RETURN_NONE;
}

#ifdef OPENCL
static PyObject* py_poissonMG_opencl_platforms(PyObject *self, PyObject *args) {
    POISSON_ERROR(poissonMG_opencl_platforms())
    Py_RETURN_NONE;  
}

static PyObject* py_poissonMG_set_opencl_platform_device(PyObject *self, PyObject *args) {
    int p,d;
    ARG_ERROR(!PyArg_ParseTuple(args,"ii",&p,&d))
    POISSON_ERROR(poissonMG_set_opencl_platform_device(p,d))
    Py_RETURN_NONE;
}
static PyObject* py_poissonMG_opencl_init(PyObject *self, PyObject *args) {
    POISSON_ERROR(poissonMG_opencl_init())
    Py_RETURN_NONE;  
}

static PyObject* py_poissonMG_opencl_create_memory_finegrid(PyObject *self, PyObject *args) {
    POISSON_ERROR(poissonMG_opencl_create_memory_finegrid())
    Py_RETURN_NONE;  
}

static PyObject* py_poissonMG_opencl_release_memory_finegrid(PyObject *self, PyObject *args) {
    POISSON_ERROR(poissonMG_opencl_release_memory_finegrid())
    Py_RETURN_NONE;  
}

static PyObject* py_poissonMG_opencl_release(PyObject *self, PyObject *args) {
    POISSON_ERROR(poissonMG_opencl_release())
    Py_RETURN_NONE;  
}

static PyObject* py_poissonMG_opencl_iterations_finegrid(PyObject *self, PyObject *args) {
    int niter;
    float w;
    ARG_ERROR(!PyArg_ParseTuple(args,"if",&niter,&w))
    POISSON_ERROR(poissonMG_opencl_iterations_finegrid(niter,w))
    Py_RETURN_NONE;
}

static PyObject* py_poissonMG_opencl_get_array_norm_finegrid(PyObject *self, PyObject *args) {
    float norm;
    POISSON_ERROR(poissonMG_opencl_get_array_norm_finegrid(&norm))
    return Py_BuildValue("f",norm);
}
#endif


static PyMethodDef pypoissonMethods[] = {
    {"open",py_poissonMG_open,METH_VARARGS},
    {"close",py_poissonMG_close,METH_NOARGS},
    {"laplacien",py_poissonMG_laplacien,METH_NOARGS},
    {"source",py_poissonMG_source,METH_VARARGS},
    {"source_polygon",py_poissonMG_source_polygon,METH_VARARGS},
    {"source_rect",py_poissonMG_source_rect,METH_VARARGS},
    {"dirichlet",py_poissonMG_dirichlet,METH_VARARGS},
    {"dirichlet_borders",py_poissonMG_dirichlet_borders,METH_VARARGS},
    {"dirichlet_polygon",py_poissonMG_dirichlet_polygon,METH_VARARGS},
    {"neumann_borders",py_poissonMG_neumann_borders,METH_VARARGS},
    {"neumann_border_ext_polar",py_poissonMG_neumann_border_ext_polar,METH_VARARGS},
    {"neumann_polygon",py_poissonMG_neumann_polygon,METH_VARARGS},
    {"neumann_border_int_polar",py_poissonMG_neumann_border_int_polar,METH_VARARGS},
    {"interface_polygon",py_poissonMG_interface_polygon,METH_VARARGS},
    {"sinus_mode",py_poissonMG_sinus_mode,METH_VARARGS},
    {"step_mode",py_poissonMG_step_mode,METH_NOARGS},
    {"init_array",py_poissonMG_init_array,METH_NOARGS},
    {"finegrid_iterations",py_poissonMG_finegrid_iterations,METH_VARARGS},
    {"finegrid_iterations_2threads",py_poissonMG_finegrid_iterations_2threads,METH_VARARGS},
    {"finegrid_iterations_4threads",py_poissonMG_finegrid_iterations_4threads,METH_VARARGS},
    {"get_array",py_poissonMG_get_array,METH_VARARGS},
    {"get_mask",py_poissonMG_get_mask,METH_VARARGS},
    {"get_derivX",py_poissonMG_get_derivX,METH_VARARGS},
    {"get_derivY",py_poissonMG_get_derivY,METH_VARARGS},
    {"get_derivRUR_axial",py_poissonMG_get_derivRUR_axial,METH_VARARGS},
    {"get_derivTheta_polar",py_poissonMG_get_derivTheta_polar,METH_NOARGS},
    {"get_RUR_axial",py_poissonMG_get_RUR_axial,METH_VARARGS},
    {"get_finegrid_residu_norm",py_poissonMG_get_finegrid_residu_norm,METH_NOARGS},
    {"get_finegrid_array_norm",py_poissonMG_get_finegrid_array_norm,METH_NOARGS},
    {"multigrid_iterations",py_poissonMG_multigrid_iterations,METH_VARARGS},
    {"multigrid_Vcycles",py_poissonMG_multigrid_Vcycles,METH_VARARGS},
    {"multigrid_full",py_poissonMG_multigrid_full,METH_VARARGS},
#ifdef OPENCL 
    {"opencl_platforms",py_poissonMG_opencl_platforms,METH_NOARGS},
    {"set_opencl_platform_device",py_poissonMG_set_opencl_platform_device,METH_VARARGS},
    {"opencl_init",py_poissonMG_opencl_init,METH_NOARGS},
    {"opencl_create_memory_finegrid",py_poissonMG_opencl_create_memory_finegrid,METH_NOARGS},
    {"opencl_release_memory_finegrid",py_poissonMG_opencl_release_memory_finegrid,METH_NOARGS},
    {"opencl_iterations_finegrid",py_poissonMG_opencl_iterations_finegrid,METH_VARARGS},
    {"opencl_release",py_poissonMG_opencl_release,METH_NOARGS},
    {"opencl_get_array_norm_finegrid",py_poissonMG_opencl_get_array_norm_finegrid,METH_NOARGS},
#endif
   {NULL,NULL} 
};

#ifdef PYTHON3
static struct PyModuleDef pypoissonMGmodule = {
    PyModuleDef_HEAD_INIT,
    "pypoissonMG", 
    0,
    -1,
    pypoissonMethods
};

PyMODINIT_FUNC
PyInit_pypoissonMG(void) {
    import_array();
    return PyModule_Create(&pypoissonMGmodule);
    
}
#else
void initpypoissonMG(void) {
    Py_InitModule("pypoissonMG",pypoissonMethods);
    import_array();
}
#endif
